"""
A module for instantiating and performing operations on Non-linear Gaussian functions.
"""

# System imports
import operator
import copy

# Third-party imports
import numpy as np

# Local imports
from veroku.factors import _factor_utils
from veroku.factors.sigma_points import get_sigma_points_array, sigma_points_array_to_joint_params
from veroku.factors._factor import Factor
from veroku.factors.gaussian import Gaussian, GaussianMixture
from veroku.factors._factor_utils import make_square_matrix, indexed_square_matrix_operation, \
    format_list_elements
from veroku.factors._factor_template import FactorTemplate

# TODO: see that lazy evaluation (recompute joint) is done sensibly and consistently


class NonLinearGaussian(Factor):
    """
    A Class for instantiating and performing operations on multivariate Gaussian mixture functions.
    """

    def __init__(self, conditioning_vars, conditional_vars, transform, noise_cov, log_weight=0.0,
                 joint_distribution=None, conditional_update_factor=None, conditioning_factor=None,
                 observed_evidence=None):
        """
        The constructor.
        :param conditioning_vars: (list) The conditioning vars.
        :param conditional_vars: (list) The conditioning vars.
        :param transform: The transformation that links the conditioning to the conditional distribution
        :param noise_cov: The noise covariance matrix of the zero mean Gaussian noise.
        :param conditional_update_factor: (factor) A factor, with same scope, subset of, the conditional vars, to be
            multiplied with the transformed (joint) distribution to form the final joint distribution.
        :param conditioning_factor: The factor with scope less than or equals to the conditional variables to be
            transformed by the transform. If the scope is less than the conditional variables set, observed evidence
            (if available) can be used to still enable the successful transformation and calculation of the joint.
        :param observed_evidence: (dict) A dictionary mapping variable names ot observed values.
        """
        if len(set(conditioning_vars).intersection(conditional_vars)) != 0:
            raise ValueError('variable overlap between conditioning_vars and conditional_vars.')
        var_names = conditioning_vars + conditional_vars

        super().__init__(var_names=var_names)
        self.transform = transform
        if noise_cov.shape[0] != len(conditional_vars):
            raise ValueError('noise_cov rows and columns should correspond to the number of conditional variables')
        self.noise_cov = make_square_matrix(noise_cov)

        self.log_weight = log_weight
        self.conditioning_vars = conditioning_vars.copy()
        self.conditional_vars = conditional_vars.copy()

        self.conditioning_factor = None
        if conditioning_factor is not None:
            self.conditioning_factor = conditioning_factor.copy()

        self.conditional_update_factor = None
        if conditional_update_factor is not None:
            self.conditional_update_factor = conditional_update_factor.copy()

        if joint_distribution is not None:
            self._joint_distribution = joint_distribution.copy()
        else:
            self._joint_distribution = Gaussian.make_vacuous(var_names)
        if observed_evidence is None:
            self.observed_evidence = {}
        else:
            self.observed_evidence = copy.deepcopy(observed_evidence)

    def copy(self):
        """
        Make a copy of this NonLinearGaussian
        :return: the copied NonLinearGaussian
        """
        nlg_copy = NonLinearGaussian(conditioning_vars=copy.deepcopy(self.conditioning_vars),
                                     conditional_vars=copy.deepcopy(self.conditional_vars),
                                     transform=copy.deepcopy(self.transform),
                                     noise_cov=copy.deepcopy(self.noise_cov),
                                     log_weight=self.log_weight,
                                     joint_distribution=self._joint_distribution,
                                     conditional_update_factor=self.conditional_update_factor,
                                     conditioning_factor=self.conditioning_factor,
                                     observed_evidence=self.observed_evidence)
        return nlg_copy

    @property
    def joint_distribution(self):
        """
        Get the joint distribution
        :return: the joint distribution
        """
        self_copy = self.copy()
        self_copy._recompute_joint()
        return self_copy._joint_distribution

    def get_K(self):
        """
        Get the K matrix (after ensuring that the canonical form is updated.)
        :return: The K parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_K()
        # pylint: enable=invalid-name

        # pylint: disable=invalid-name

    def get_h(self):
        """
        Get the log h vector (after ensuring that the canonical form is updated.)
        :return: The h parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_h()
        # pylint: enable=invalid-name

        # pylint: disable=invalid-name

    def get_g(self):
        """
        Get the g scalar (after ensuring that the canonical form is updated.)
        :return: The g parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_g()
        # pylint: enable=invalid-name

        # pylint: disable=invalid-name

    def get_cov(self):
        """
        Get the covariance matrix (after ensuring that the covariance form is updated.)
        :return: The cov parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_cov()
        # pylint: enable=invalid-name

    def get_mean(self):
        """
        Get the mean vector (after ensuring that the covariance form is updated.)
        :return: The mean parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_mean()

    def get_log_weight(self):
        """
        Get the log weight (after ensuring that the covariance form is updated.)
        :return: The log_weight parameter.
        """
        self._recompute_joint()
        return self._joint_distribution.get_log_weight()

    def add_log_weight(self, log_weight_to_add):
        self.log_weight += log_weight_to_add

    def _recompute_joint(self):
        observed_vars = list(self.observed_evidence.keys())

        conditioning_observed_vars = [var for var in observed_vars if var in self.conditioning_vars]
        conditioning_observed_values = [self.observed_evidence[var] for var in observed_vars if
                                        var in self.conditioning_vars]

        conditional_observed_vars = [var for var in observed_vars if var in self.conditional_vars]
        conditional_observed_values = [self.observed_evidence[var] for var in observed_vars if
                                       var in self.conditional_vars]

        # under 3 conditions we can successfully perform the transformation:
        # 1. We have a well defined conditional distribution with the full scope of the conditional variables.
        # 2. All the conditional variables are observed.
        # 3. We have a conditional distribution over a subset of the conditional variables and the rest of the
        # conditional variables are observed.

        # Initial feasibility check
        conditioning_factor_var_names = []
        if self.conditioning_factor is not None:
            conditioning_factor_var_names = self.conditioning_factor.var_names
        if set(conditioning_factor_var_names).union(set(conditioning_observed_vars)) != set(
                self.conditioning_vars):
            # The conditioning_factor variables and observed conditional variables combined do not make up a
            # complete set - so we cannot calculate the transform.
            self._joint_distribution = Gaussian.make_vacuous(self.var_names)
            return

        # If we are still here, there we can calculate the transformation.
        conditioning_factor_sigma_points_array = np.expand_dims(np.array([]), axis=0).T  # empty dummy for generalisation
        if self.conditioning_factor is not None:
            conditioning_factor = self.conditioning_factor.copy()
            cond_factor_observed_vars = list(set(observed_vars).intersection(conditioning_factor.var_names))
            cond_factor_observed_values = [self.observed_evidence[var] for var in cond_factor_observed_vars]

            if len(cond_factor_observed_vars) > 0:
                conditioning_factor = conditioning_factor.reduce(vrs=cond_factor_observed_vars,
                                                                 values=cond_factor_observed_values)

            conditioning_factor_sigma_points_array = get_sigma_points_array(conditioning_factor)

        # extend sigma points with observed variable values
        if len(conditioning_observed_vars) > 0:
            observed_vec = _factor_utils.make_column_vector(conditioning_observed_values)
            tiled_observed_vec = np.tile(observed_vec, [1, conditioning_factor_sigma_points_array.shape[1]])
            extended_sigma_points_array = np.concatenate([tiled_observed_vec,
                                                          conditioning_factor_sigma_points_array])
            extended_sigma_point_vars = conditioning_observed_vars + conditioning_factor.var_names
        else:
            extended_sigma_points_array = conditioning_factor_sigma_points_array
            extended_sigma_point_vars = conditioning_factor.var_names

        # TODO: Add conditional var_names here somehow. We need to ensure that the correct variables are used
        # in the correct place in the transformation. In the past, we have simply done this by standardising
        # on the variable indices in different place, but this is probably not very safe.
        joint_cov, joint_mean = sigma_points_array_to_joint_params(sigma_points_array=extended_sigma_points_array,
                                                                   transform=self.transform,
                                                                   var_names=extended_sigma_point_vars)
        if (joint_mean.shape[0] != len(self.var_names)) or (joint_cov.shape[0] != len(self.var_names)):
            raise AssertionError('Transform resulted in incorrect number of variables.')

        # remove the observed vars again after transformation
        nov = len(conditioning_observed_vars)  # number of observed vars
        joint_cov = joint_cov[nov:, nov:]
        joint_mean = joint_mean[nov:, :]
        joint_var_names = conditioning_factor.var_names + self.conditional_vars

        joint_cov_plus_noise, joint_vars = indexed_square_matrix_operation(joint_cov, self.noise_cov,
                                                                           joint_var_names,
                                                                           self.conditional_vars,
                                                                           operator.add)
        self._joint_distribution = Gaussian(cov=joint_cov_plus_noise,
                                            mean=joint_mean,
                                            log_weight=self.log_weight + conditioning_factor.log_weight,
                                            var_names=joint_vars)
        if self.conditional_update_factor is not None:
            self._joint_distribution = self._joint_distribution.multiply(self.conditional_update_factor)
        if len(conditional_observed_vars) > 0:
            self._joint_distribution = self._joint_distribution.reduce(conditional_observed_vars,
                                                                       conditional_observed_values)

    def multiply(self, factor):
        """
        Multiply this factor with another factor.
        :param factor: the factor to multiply with
        :return: the resulting GaussianMixture
        """

        if isinstance(factor, GaussianMixture):
            nlgs = []
            for gaussian_factor in factor.components:
                nlgs.append(self.multiply(gaussian_factor))
            return NonLinearGaussianMixture(nlgs)

        self_copy = self.copy()
        if isinstance(factor, NonLinearGaussian):
            raise NotImplementedError('Multiplication of two nonlinear-Gaussian factors has not been implemented.')
        if factor._is_vacuous:
            # TODO: check this (esp to see what to do with)
            return self_copy
        # TODO: change to 'update conditional model and re-transform model' (if correct)
        if not isinstance(factor, Gaussian):
            raise ValueError(f'factor must be Gaussian, got {type(factor)}')

        if set(factor.var_names) <= set(self_copy.conditional_vars):
            if self_copy.conditional_update_factor is None:
                self_copy.conditional_update_factor = factor.copy()
            else:
                self_copy.conditional_update_factor = self_copy.conditional_update_factor.multiply(factor)
        elif set(factor.var_names) <= set(self_copy.conditioning_vars):
            if self_copy.conditioning_factor is None:
                self_copy.conditioning_factor = factor.copy()
            else:
                self_copy.conditioning_factor = self_copy.conditioning_factor.multiply(factor)
        else:
            raise ValueError(f'cannot absorb factor with scope ({factor.var_names}) \n '
                             f'           which has neither conditional ({self_copy.conditioning_vars}) \n '
                             f'           nor conditioning ({self_copy.conditional_vars}) scope.')
        return self_copy

    def divide(self, factor):
        """
        Divide this factor by another factor.
        :param factor: the factor divide by
        :return: the resulting GaussianMixture
        """
        self_copy = self.copy()
        if factor._is_vacuous:
            # TODO: check this (esp to see what to do with)
            return self_copy
        # TODO: change to 'update conditional model and re-transform model' (if correct)
        if not isinstance(factor, Gaussian):
            raise ValueError('factor must be Gaussian.')

        if set(factor.var_names) <= set(self_copy.conditional_vars):
            self_copy.conditional_update_factor = self_copy.conditional_update_factor.divide(factor)
        elif set(factor.var_names) <= set(self_copy.conditioning_vars):
            if self_copy.conditioning_factor is None:
                raise NotImplementedError()
            self_copy.conditioning_factor = self_copy.conditioning_factor.divide(factor)
        else:
            raise ValueError(f'cannot cancel factor with scope ({factor.var_names}) \n '
                             f'           which has neither conditional ({self_copy.conditioning_vars}) \n '
                             f'           nor conditioning ({self_copy.conditional_vars}) scope.')
        self_copy._recompute_joint()
        return self_copy

    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this factor and return the resulting factor.
        :param vrs: the names of the observed variable (list)
        :param values: the values of the observed variables (list or vector-like object)
        :return: the resulting GaussianMixture
        """

        self_copy = self.copy()
        self_copy.observed_evidence.update({v: val for v, val in zip(vrs, values)})
        return self_copy

    def marginalize(self, vrs, keep=False):
        """
        Integrate out variables from this factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: the resulting Gaussian mixture marginal
        """

        vrs_to_keep = super().get_marginal_vars(vrs, keep=keep)
        assert set(vrs_to_keep) <= set(self.var_names), 'Error: asked to keep vars not in factor.'
        self_copy = self.copy()
        self_copy._recompute_joint()
        if self_copy._joint_distribution._is_vacuous:
            if set(vrs_to_keep) <= set(self_copy.conditioning_vars):
                if self_copy.conditioning_factor is not None:
                    if set(vrs_to_keep) <= set(self_copy.conditioning_factor.var_names):
                        marginal = self_copy.conditioning_factor.marginalize(vrs_to_keep, keep=True)
                        return marginal
            if set(vrs_to_keep) <= set(self_copy.conditional_vars):
                if self_copy.conditional_update_factor is not None:
                    if set(vrs_to_keep) <= set(self_copy.conditional_update_factor.var_names):
                        marginal = self_copy.conditional_update_factor.marginalize(vrs_to_keep, keep=True)
                        marginal._add_log_weight(self.log_weight)
                        return marginal
            else:
                return Gaussian.make_vacuous(vrs_to_keep)
        return self_copy._joint_distribution.marginalize(vrs_to_keep, keep=True)

    def sample(self, num_samples):
        self._recompute_joint()
        return self._joint_distribution.sample(num_samples=num_samples)

    def show(self):
        """
        Print the parameters of the nonlinear Gaussian distribution
        """
        print('Non-linear Gaussian')
        print('conditioning variables:', self.conditioning_vars)
        print('conditional variables:', self.conditional_vars)
        print('transform: ', self.transform)
        print('noise covariance = \n', self.noise_cov)
        print('joint_distribution = \n')
        if self._joint_distribution is not None:
            self._recompute_joint()
            self._joint_distribution.show()

        print('conditional update distribution = \n')
        if self.conditional_update_factor is not None:
            self.conditional_update_factor.show()
        else:
            print('vacuous')

        print('conditioning distribution = \n')
        if self.conditioning_factor is not None:
            self.conditioning_factor.show()
        else:
            print('vacuous')

    def plot(self):
        self_copy = self.copy()
        self_copy._recompute_joint()
        self._joint_distribution.plot()


class NonLinearGaussianMixture(Factor):
    """
    A Class for instantiating and performing operations on multivariate Gaussian mixture functions.
    """

    def __init__(self, factors):
        """
        The constructor.
        :param factors: (NonLinearGaussian factor list) The list of factors.
        """
        self.nlgs = []
        if len(factors) == 0:
            raise ValueError('received empty list of factors.')
        self._var_names = factors[0].var_names
        for factor in factors:
            if not isinstance(factor, NonLinearGaussian):
                raise TypeError(f'expected NonLinearGaussian type, received {type(factor)}.')
            if not set(self.var_names) == set(factor.var_names):
                raise ValueError(f'Inconsistent variable names. First factor has var_names = {self.var_names},'
                                 f' another has var_names = {factor.var_names}')
            self.nlgs.append(factor.copy())

    def copy(self):
        """
        Make a copy of this NonLinearGaussianMixture
        :return: the copied NonLinearGaussianMixture
        """
        return NonLinearGaussianMixture(self.nlgs)

    @property
    def _joint_distribution(self):
        gaussian_joints = []
        for nlg in self.nlgs:
            gaussian_joints.append(nlg._joint_distribution)
        return GaussianMixture(gaussian_joints)

    def multiply(self, factor):
        """
        Multiply this factor with another factor.
        :param factor: the factor to multiply with
        :return: the resulting GaussianMixture
        """
        if isinstance(factor, GaussianMixture):
            gm_factor = factor
        elif isinstance(factor, Gaussian):
            gm_factor = GaussianMixture([Gaussian])
        else:
            raise NotImplementedError()
        new_nlgs = []
        for gauss in gm_factor.components:
            for nlg in self.nlgs:
                new_nlgs.append(nlg.multiply(gauss))
        return NonLinearGaussianMixture(new_nlgs)

    def divide(self, factor):
        """
        Divide this factor by another factor.
        :param factor: the factor divide by
        :return: the resulting GaussianMixture
        """
        if isinstance(factor, Gaussian):
            gaussian_factor = factor
        elif isinstance(factor, GaussianMixture):
            # TODO: Add better Gaussian mixture division approximation
            gaussian_factor = factor.moment_match_to_single_gaussian()
        else:
            raise NotImplementedError('cannot divide NonLinearGaussianMixture by {type(factor)}.')
        new_nlgs = []
        for nlg in self.nlgs:
            new_nlgs.append(nlg.divide(gaussian_factor))
        return NonLinearGaussianMixture(new_nlgs)

    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this factor and return the resulting factor.
        :param vrs: the names of the observed variable (list)
        :param values: the values of the observed variables (list or vector-like object)
        :return: the resulting GaussianMixture
        """
        new_nlgs = []
        for nlg in self.nlgs:
            new_nlgs.append(nlg.reduce(vrs, values))
        return NonLinearGaussianMixture(new_nlgs)

    def marginalize(self, vrs, keep=False):
        """
        Integrate out variables from this factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: the resulting Gaussian mixture marginal
        """
        new_nlgs = []
        for nlg in self.nlgs:
            new_nlgs.append(nlg.marginalize(vrs, keep))
        return NonLinearGaussianMixture(new_nlgs)

    def sample(self, num_samples):
        #TODO: fix incorrect num samples issue
        sample_sets = []
        for nlg in self.nlgs:
            samples = nlg.sample(num_samples)
            sample_sets.append(samples)
        return np.concatenate(sample_sets)

    def plot(self):
        """
        Plot the joint Gaussian Mixture distributio
        :return:
        """
        gaussian_components = [nlg.joint_distribution for nlg in self.nlgs]
        gaussian_mixture = GaussianMixture(gaussian_components)
        gaussian_mixture.plot()

    def show(self):
        """
        Print the parameters of the nonlinear Gaussian distribution
        """
        for i, nlg in enumerate(self.nlgs):
            print(f'######### NLG {i} #############################')
            nlg.show()

    class NonLinearGaussianMixtureTemplate(FactorTemplate):

        def __init__(self, conditioning_var_templates, conditional_var_templates, transition_function, noise_cov):
            """
            transition_function()
            """
            super().__init__(conditioning_var_templates=conditioning_var_templates,
                             conditional_var_templates=conditional_var_templates)
            self.transition_function = transition_function
            self.noise_cov = noise_cov

        def make_factor(self, format_dict=None, var_names=None):
            """
            Make a factor with var_templates formatted by format_dict to create specific var names.
            :param format_dict:
            :return:
            """
            if format_dict is not None:
                assert var_names is None
                #  TODO: use format_list_elements
                conditioning_vars = [vt.format(**format_dict) for vt in self._conditioning_var_templates]
                conditional_vars = [vt.format(**format_dict) for vt in self._conditional_var_templates]

                return NonLinearGaussianMixture(conditioning_vars, conditional_vars,
                                                self.transition_function, self.noise_cov,
                                                log_weight=0.0,
                                                joint_distribution=None)


class NonLinearGaussianTemplate(FactorTemplate):

    def __init__(self, conditioning_var_templates, conditional_var_templates, transition_function, noise_cov):
        """
        The initialiser.

        :param conditioning_var_templates:
        :param conditional_var_templates:
        :param transition_function:
        :param noise_cov:
        """
        super().__init__(conditioning_var_templates=conditioning_var_templates,
                         conditional_var_templates=conditional_var_templates)
        self.transition_function = transition_function
        self.noise_cov = noise_cov

    def make_factor(self, format_dict=None, var_names=None):
        """
        Make a factor with var_templates formatted by format_dict to create specific var names.
        :param format_dict:
        :return:
        """
        if format_dict is not None:
            assert var_names is None

            conditioning_vars = format_list_elements(self._conditioning_var_templates, format_dict)
            conditional_vars = format_list_elements(self._conditional_var_templates, format_dict)

            return NonLinearGaussian(conditioning_vars, conditional_vars,
                                     self.transition_function, self.noise_cov,
                                     log_weight=0.0,
                                     joint_distribution=None)



from IPython.core.display import Image, display
from graphviz import Graph
import networkx as nx
import numpy as np
from tqdm.auto import tqdm
import pandas as pd
from graphviz import Source

from veroku._cg_helpers._cluster import Cluster
import veroku._cg_helpers._animation as cg_animation
from veroku.factors._factor_utils import get_subset_evidence
import matplotlib.pyplot as plt
import collections

# TODO: optimise _pass_message
# TODO: improve sepsets selection for less loopiness
# TODO: Optimisation: messages from clusters that did not receive any new messages in the previous round, do not need
#  new messages calculated.


# TODO: consider removing this
def sort_almost_sorted(a_deque, key):
    """
    Sort a deque like that where only the first element is potentially unsorted
    and should probably be last and the rest of the deque is sorted in descending order.
    """
    a_deque.append(a_deque.popleft())
    if key(a_deque[-1]) <= key(a_deque[-2]):
        return a_deque
    a_deque = collections.deque(sorted(a_deque, key=key, reverse=True))
    return a_deque


def _evidence_reduce_factors(factors, evidence):
    """
    Observe relevant evidence for each factor.
    :param factors:
    :param evidence:
    :return:
    """
    reduced_factors = []
    for i, factor in enumerate(factors):
        if evidence is not None:
            vrs, values = get_subset_evidence(all_evidence_dict=evidence,
                                              subset_vars=factor.var_names)
            if len(vrs) > 0:
                factor = factor.reduce(vrs, values)
        reduced_factors.append(factor.copy())
    return reduced_factors


def make_factor_name(factor):
    return str(factor.var_names).replace("'", '')


def _absorb_subset_factors(factors):
    """
    Absorb any factors that has a scope that is a subset of another factor into such a factor.
    :param factors:
    :return:
    """
    factors_absorbtion_dict = {i: [] for i in range(len(factors))}
    final_graph_cluster_factors = []
    # factors: possibly smaller list of factors after factors which have a scope that is a subset of another factor have
    # been absorbed by the larger one.
    factor_processed_mask = [0] * len(factors)
    for i, factor_i in enumerate(factors):
        if not factor_processed_mask[i]:
            factor_product = factor_i.copy()
            for j, factor_j in enumerate(factors):
                if i != j:
                    if not factor_processed_mask[j]:
                        if set(factor_j.var_names) < set(factor_product.var_names):
                            try:
                                factor_product = factor_product.multiply(factor_j)
                                factors_absorbtion_dict[i].append(j)
                                factor_processed_mask[j] = 1
                                factor_processed_mask[i] = 1
                            except NotImplementedError:
                                print(f'Warning: could not multiply {type(factor_product)} with {type(factor_j)} (Not Implemented)')
            if factor_processed_mask[i]:
                final_graph_cluster_factors.append(factor_product)
    for i, factor_i in enumerate(factors):  # add remaining factors
        if not factor_processed_mask[i]:
            factor_processed_mask[i] = 1
            final_graph_cluster_factors.append(factor_i)
    assert all(factor_processed_mask), 'Error: Some factors where not included during variable subset processing.'
    return final_graph_cluster_factors


def _make_subset_factor_df(subset_dict):
    """
    Make a ...
    :param subset_dict: (dict) A dictionary mapping factors to factors that have subset scopes
    :return:
    """
    keys = list(subset_dict.keys())
    values = list(subset_dict.values())
    assert(len(keys) == len(values))
    # TODO: This raises different list lengths warning (see below). Investigate this.
    #   VisibleDeprecationWarning: Creating an ndarray from ragged nested sequences...
    data = np.array([keys, values]).T
    df = pd.DataFrame(columns=['factor_index', 'subfactor_indices'],
                      data=data)
    df['num_subfactors'] = df['subfactor_indices'].apply(lambda x: len(x))
    df.sort_values(by='num_subfactors', inplace=True, ascending=False)
    return df


class ClusterGraph(object):
    
    def __init__(self, factors, evidence=None, special_evidence=dict(),
                 make_animation_gif=False, disable_tqdm=False, verbose=False):
        """
        Construct a Cluster graph from a list of factors
        :param factors: (list of factors) The factors to construct the graph from
        :param evidence: (dict) evidence dictionary (mapping variable names to values) that should be used to reduce
            factors before building the cluster graph.
        :param special_evidence: (dict) evidence dictionary (mapping variable names to values) that should be used in
            the calculation of messages, and not to reduce factors. This allows factor approximations - such as the
            non-linear Gaussian to be iteratively refined.
        """
        # TODO: see if evidence and special_evidence can be replaced by a single variable.
        self.num_messages_passed = 0
        self.make_animation_gif = make_animation_gif
        self.special_evidence = special_evidence
        self.disable_tqdm = disable_tqdm
        self.last_passed_message_factors_dict = dict()
        self.verbose = verbose
        # new
        self.last_sent_message_dict = {}  # {(rec_cluster_id1, rec_cluster_id1): msg1, ...}

        self.sync_message_passing_max_distances = []

        all_evidence_vars = set(self.special_evidence.keys())
        if evidence is not None:
            evidence_vars = set(evidence.keys())
            all_evidence_vars = all_evidence_vars.union(evidence_vars)
        all_factors_copy = _evidence_reduce_factors(factors, evidence)
        final_graph_cluster_factors = _absorb_subset_factors(all_factors_copy)

        clusters = [Cluster(factor, cluster_name_prefix=f'c{i}#') for i, factor in
                    enumerate(final_graph_cluster_factors)]

        self._set_non_rip_sepsets_dict(clusters, all_evidence_vars)
        self._clusters = clusters

        # Add special evidence to factors
        for cluster in self._clusters:
            cluster_special_evidence_vars, cluster_special_evidence_values = get_subset_evidence(self.special_evidence, cluster.var_names)
            cluster_special_evidence = dict(zip(cluster_special_evidence_vars, cluster_special_evidence_values))
            cluster.add_special_evidence(cluster_special_evidence)

        self.graph_message_paths = collections.deque([])
        self._build_graph()

        # TODO: consolidate these two, if possible
        self.message_passing_log_df = None
        self.message_passing_animation_frames = []

    def _set_non_rip_sepsets_dict(self, clusters, all_evidence_vars):
        """
        Calculate the preliminary sepsets dict before the RIP property is enforced.
        :param clusters:
        :param all_evidence_vars:
        :return:
        """
        self._non_rip_sepsets = {}
        for i in tqdm(range(len(clusters)), disable=self.disable_tqdm):
            vars_i = clusters[i].var_names
            for j in range(i + 1, len(clusters)):
                vars_j = clusters[j].var_names
                sepset = set(vars_j).intersection(set(vars_i)) - all_evidence_vars
                self._non_rip_sepsets[(i, j)] = sepset
                self._non_rip_sepsets[(j, i)] = sepset

    def _build_graph(self):
        """
        Add the cluster sepsets, graphviz graph and animation graph (for message_passing visualisation).
        """
        # Check for non-unique cluster_ids (This should never be the case)
        cluster_ids = [cluster.cluster_id for cluster in self._clusters]
        if len(set(cluster_ids)) != len(cluster_ids):
            raise ValueError(f'Non-unique cluster ids: {cluster_ids}')

        self._conditional_print('Info: Building graph.')
        self._graph = Graph(format='png')
        rip_sepsets_dict = self._get_running_intersection_sepsets()

        # TODO: see why this is necessary, remove if not
        for i in tqdm(range(len(self._clusters)), disable=self.disable_tqdm):
            self._clusters[i].remove_all_neighbours()

        self._conditional_print(f'Debug: number of clusters: {len(self._clusters)}')
        for i in tqdm(range(len(self._clusters)), disable=self.disable_tqdm):

            node_i_name = self._clusters[i]._cluster_id
            self._graph.node(name=node_i_name, label=node_i_name, style='filled', fillcolor='white', color='black')
            for j in range(i + 1, len(self._clusters)):

                if (i, j) in rip_sepsets_dict:
                    sepset = rip_sepsets_dict[(i, j)]
                    assert len(sepset) > 0, 'Error: empty sepset'
                    self._clusters[i].add_neighbour(self._clusters[j], sepset=sepset)
                    self._clusters[j].add_neighbour(self._clusters[i], sepset=sepset)

                    gmp_ij = _GraphMessagePath(self._clusters[i], self._clusters[j])
                    gmp_ji = _GraphMessagePath(self._clusters[j], self._clusters[i])
                    self.graph_message_paths.append(gmp_ij)
                    self.graph_message_paths.append(gmp_ji)
                    self._clusters[i].add_outward_message_path(gmp_ij)
                    self._clusters[j].add_outward_message_path(gmp_ji)

                    # Graph animation
                    node_j_name = self._clusters[j]._cluster_id
                    sepset_node_label = ','.join(sepset)
                    sepset_node_name = cg_animation.make_sepset_node_name(node_i_name, node_j_name)
                    self._graph.node(name=sepset_node_name, label=sepset_node_label, shape='rectangle')
                    self._graph.edge(node_i_name, sepset_node_name, color='black', penwidth='2.0')
                    self._graph.edge(sepset_node_name, node_j_name, color='black', penwidth='2.0')
        print('num self.graph_message_paths: ', len(self.graph_message_paths))

    def _conditional_print(self, message):
        if self.verbose:
            print(message)

    def plot_next_messages_info_gain(self, legend_on=False, figsize=[15, 5]):
        """
        Plot the information gained by a receiving new messages over sebsequent iterations for all message paths in the
            graph.
        :param bool legend_on: Whether or not to show the message paths (specified by connected cluster pairs) in the
            plot legend.
        :param list figsize: The matplotlib figure size.
        """
        plt.figure(figsize=figsize)
        all_paths_information_gains_with_iters = [gmp.information_gains_with_iters for gmp in self.graph_message_paths]
        for paths_information_gains_with_iters in all_paths_information_gains_with_iters:
            plt.plot(paths_information_gains_with_iters)
        plt.title('Information Gain of Messages along Graph Message Paths')
        plt.xlabel('iteration')
        plt.ylabel('D_KL(prev_msg||msg)')
        if legend_on:
            legend = [f"{gmp.sender_cluster.cluster_id}->{gmp.receiver_cluster.cluster_id}" for gmp in
                      self.graph_message_paths]
            plt.legend(legend)

    def plot_message_convergence(self, log=False, figsize=[15, 5]):
        """
        Plot the the KL-divergence between the messages and their previous instances to indicate the message passing
        convergence.
        """

        mp_max_dists = self.sync_message_passing_max_distances
        if log:
            mp_max_dists = np.log(mp_max_dists)
        from matplotlib.lines import Line2D
        # here we tile an flatten to prevent the plot omission of values with inf on either side.
        mp_max_dists = np.tile(mp_max_dists, [2, 1]).flatten(order='F')
        num_iterations = len(mp_max_dists)

        iterations = np.array(list(range(num_iterations))) / 2  # divide by 2 to correct for tile and flatten
        non_inf_max_distances = [d for d in mp_max_dists if d != np.inf]
        max_non_inf = max(non_inf_max_distances)
        new_inf_value = max_non_inf * 1.5
        max_distances_replaces_infs = np.array([v if v != np.inf else new_inf_value for v in mp_max_dists])
        inf_values = np.ma.masked_where(max_distances_replaces_infs != new_inf_value, max_distances_replaces_infs)
        plt.figure(figsize=figsize)
        plt.plot(iterations, max_distances_replaces_infs)
        plt.plot(iterations, inf_values, c='r', linewidth=2)
        if len(non_inf_max_distances) != len(mp_max_dists):
            custom_lines = [Line2D([0], [0], color='r', lw=4)]
            plt.legend(custom_lines, ['infinity'])
        plt.title('Message Passing Convergence')
        plt.xlabel('iteration')
        plt.ylabel('log max D_KL(prev_msg||msg)')
        plt.show()

    def _get_unique_vars(self):
        all_vars = []
        for cluster in self._clusters:
            all_vars += (cluster.var_names)
        unique_vars = list(set(all_vars))
        return unique_vars

    def _get_vars_min_spanning_trees(self):
        all_vars = self._get_unique_vars()
        var_graphs = {var: nx.Graph() for var in all_vars}
        num_clusters = len(self._clusters)
        for i in range(num_clusters):
            for j in range(i + 1, num_clusters):
                sepset = self._non_rip_sepsets[(i, j)]
                for var in sepset:
                    var_graphs[var].add_edge(i, j, weight=1)
        var_spanning_trees = dict()
        for var in all_vars:
            var_graph = var_graphs[var]
            var_spanning_trees[var] = nx.minimum_spanning_tree(var_graph)
        return var_spanning_trees

    def _get_running_intersection_sepsets(self):
        edge_sepset_dict = {}
        unique_vars = self._get_unique_vars()
        min_span_trees = self._get_vars_min_spanning_trees()
        self._conditional_print("Info: Getting unique variable spanning trees.")
        for i in tqdm(range(len(unique_vars)), disable=self.disable_tqdm):
            var = unique_vars[i]
            min_span_tree = min_span_trees[var]
            for edge in min_span_tree.edges():
                if edge in edge_sepset_dict:
                    edge_sepset_dict[edge].append(var)
                else:
                    edge_sepset_dict[edge] = [var]
        return edge_sepset_dict

    def show(self):
        """
        Show the cluster graph.
        """
        self._graph.render('/tmp/test.gv', view=False)
        image = Image('/tmp/test.gv.png')
        display(image)

    def save_graph_image(self, filename):
        """
        Save image of the graph.
        :param filename: The filename of the file.
        """
        # Source(self._graph, filename="/tmp/test.gv", format="png")
        Source(self._graph, filename=filename, format="png")

    def get_marginal(self, vrs):
        """
        Search the graph for a specific variable and get that variables marginal (posterior marginal if process_graph
        has been run previously).
        :return: The marginal
        """
        for cluster in self._clusters:
            if set(vrs) <= set(cluster.var_names):
                factor = cluster._factor.copy()
                evidence_vrs, evidence_values = get_subset_evidence(self.special_evidence, factor.var_names)
                if len(evidence_vrs) > 0:
                    factor = factor.reduce(evidence_vrs, evidence_values)
                marginal = factor.marginalize(vrs, keep=True)
                return marginal
        raise ValueError(f'No cluster with variables containing {vrs}')

    def get_posterior_joint(self):
        """
        Get the posterior joint distribution.
        """
        # TODO: add functionality for efficiently getting a posterior marginal over any subset of variables and replace
        #  the get_marginal function above.
        cluster_product = self._clusters[0]._factor.joint_distribution
        for cluster in self._clusters[1:]:
            cluster_product = cluster_product.multiply(cluster._factor.joint_distribution)
        last_passed_message_factors = list(self.last_passed_message_factors_dict.values())
        if len(last_passed_message_factors) == 0:
            assert self.num_messages_passed == 0
            return cluster_product
        message_product = last_passed_message_factors[0]
        for message_factor in last_passed_message_factors[1:]:
            message_product = message_product.multiply(message_factor)
        joint = cluster_product.cancel(message_product)
        return joint

    def process_graph(self, tol=1e-3, max_iter=50, debug=False):
        """
        Perform synchronous message passing until convergence (or maximum iterations).
        """

        # for debugging and testing
        self.messages_passed = []

        self.sync_message_passing_max_distances = []
        if len(self._clusters) == 1:
            # The Cluster Graph contains only single cluster. Message passing not possible or necessary.
            self._clusters[0]._factor = self._clusters[0]._factor.reduce(vrs=self.special_evidence.keys(),
                                                                         values=self.special_evidence.values())
            return

        # TODO: see if the definition of max_iter can be improved
        key_func = lambda x: x.next_information_gain
        max_message_passes = max_iter*len(self.graph_message_paths)

        self.graph_message_paths = collections.deque(sorted(self.graph_message_paths, key=key_func, reverse=True))

        for _ in tqdm(range(max_message_passes), disable=self.disable_tqdm):
            sender_cluster_id = self.graph_message_paths[0].sender_cluster.cluster_id
            receiver_cluster_id = self.graph_message_paths[0].receiver_cluster.cluster_id

            # for debugging and testing
            if debug:
                self.messages_passed.append(self.graph_message_paths[0].next_message.copy())
                data = [(gmp.sender_cluster._cluster_id, gmp.receiver_cluster._cluster_id, gmp.next_information_gain) for
                        gmp in self.graph_message_paths]
                df = pd.DataFrame(data=data, columns=['sender', 'receiver', 'info'])

            self.graph_message_paths[0].pass_next_message()
            self.graph_message_paths = collections.deque(sorted(self.graph_message_paths, key=key_func, reverse=True))
            # self.graph_message_paths = sort_almost_sorted(self.graph_message_paths, key=key_func)

            max_next_information_gain = self.graph_message_paths[0].next_information_gain
            self.sync_message_passing_max_distances.append(max_next_information_gain)
            if max_next_information_gain <= tol:
                return

            if self.make_animation_gif:
                cg_animation.add_message_pass_animation_frames(graph=self._graph,
                                                               frames=self.message_passing_animation_frames,
                                                               node_a_name=sender_cluster_id,
                                                               node_b_name=receiver_cluster_id)

    def _make_message_passing_animation_gif(self):
        print('Making message passing animation.')
        self.message_passing_animation_frames[0].save(fp='./graph_animation.gif',
                                                      format='GIF',
                                                      append_images=self.message_passing_animation_frames[1:],
                                                      save_all=True, duration=400, loop=0)


class _GraphMessagePath:
    """
    A specific path (direction along an edge) in a graph along which a message can be passed.
    """
    def __init__(self, sender_cluster, receiver_cluster):
        """
        The initialiser.
        :param sender_cluster: The cluster that defines the starting point of the path.
        :param receiver_cluster: The cluster that defines the end point of the path.
        """
        self.sender_cluster = sender_cluster
        self.receiver_cluster = receiver_cluster
        self.previously_sent_message = None
        self.next_message = self.sender_cluster.make_message(self.receiver_cluster._cluster_id)
        self.next_information_gain = None
        self.information_gains_with_iters = []
        self.update_next_information_gain()

    def update_next_information_gain(self):
        if self.previously_sent_message is None:
            self.next_information_gain = self.next_message.distance_from_vacuous()
        else:
            # "In the context of machine learning, KL(P||Q) is often called the information gain achieved if Q is
            # used instead of P." - wikipedia
            # We typically want to know which new message (Q) will result in the largest information gain if it replaces
            # the message (P)
            # message: previous_message (P)
            # factor: next message (Q)
            # P.kl_divergence(Q)
            self.next_information_gain = self.previously_sent_message.kl_divergence(self.next_message)
        self.information_gains_with_iters.append(self.next_information_gain)

    def recompute_next_message(self):
        """
        Recompute the next message.
        """
        new_next_message = self.sender_cluster.make_message(self.receiver_cluster._cluster_id)
        self.next_message = new_next_message.copy()
        self.update_next_information_gain()

    def pass_next_message(self):
        """
        Pass the next message along this path.
        """
        self.receiver_cluster.receive_message(self.next_message)
        # we 'recompute' the next message - but it will be the same
        self.previously_sent_message = self.next_message.copy()
        self.next_information_gain = 0.0
        self.information_gains_with_iters.append(self.next_information_gain)

        for gmp in self.receiver_cluster._outward_message_paths:
            gmp.recompute_next_message()



"""
A module for instantiating sparse tables with log probabilities.
"""

# System imports
import copy
import operator
import time

# Third-party imports
import numpy as np
from scipy import special
import pandas as pd
import itertools

# Local imports
from veroku.factors._factor import Factor
from veroku.factors._factor_template import FactorTemplate


def _make_dense(factor):
    """
    Convert a factor to a dense factor by instanciating all of the implicit default value assignments.
    :param factor: The factor to convert.
    :type factor: SparseCategorical
    :return: The dense factor version.
    :rtype: SparseCategorical
    """
    factor_copy = factor.copy()
    dense_assingments = itertools.product(*[range(factor_copy.var_cards[v]) for v in factor_copy.var_names])
    for assign in dense_assingments:
        if assign not in factor_copy.log_probs_table:
            factor_copy.log_probs_table[assign] = factor_copy.default_log_prob
    return factor_copy


def _make_dense_default_probs_dict(cardinalities, default_value):
    """
    Make a dictionary representing a categorical probability table that contains all assignments, as specified by the
    given cardinalities (starting from 0).
    :param cardinalities: The given cardinalities.
    :type cardinalities: list
    :param default_value: The default value to assign the missing values.
    :type default_value: float
    :returns: The dense representation.
    :rtype: dict
    Example:
    >>> _make_dense_default_probs_dict(cardinalities=[2,3], default_value=0.0)
    {(0,0):0.0,
    (0,1):0.0,
    (0,2):0.0,
    (1,0):0.0,
    (1,1):0.0,
    (1,2):0.0}
    """
    dense_assingments = itertools.product(*[range(c) for c in cardinalities])
    dense_default_dict = dict(zip(dense_assingments, [default_value] * np.product(cardinalities)))
    return dense_default_dict


def _make_inner_dense(sparse_nested_table, outer_inner_cards, default_value):
    """
    Convert n sparse nested table dictionary's implicit default values in the innetables to real entries so that
    all possible assignments are present in the inner sub tables.
    :param sparse_nested_table: The nested dictionary representing a sparse table.
    :type sparse_nested_table: dictionary of dictionaries
    :param outer_inner_cards: The lists of the cardinalities for the outer and inner part of the nested table respectively
        i.e [[2,2], [3, 4]] if the table dict is {(0,0):{(2,3):0.3},....} where the outer variables has cardinalities [2,2]
        and the inner ones [3,4]
    """
    sparse_nested_inner_dense_table = copy.deepcopy(sparse_nested_table)
    inner_cards = outer_inner_cards[1]

    all_inner_table_assignments = itertools.product(*[range(c) for c in inner_cards])
    for outer_assign in sparse_nested_inner_dense_table.keys():
        for inner_assign in all_inner_table_assignments:
            if inner_assign not in sparse_nested_inner_dense_table[outer_assign]:
                sparse_nested_inner_dense_table[outer_assign][inner_assign] = default_value
    return sparse_nested_inner_dense_table


def _get_nested_sorted_probs(new_variables_order_outer,
                             new_variables_order_inner,
                             old_variable_order, old_assign_probs):
    """
    Reorder probs to a new order and sort assignments.
    :params new_variables_order_outer:
    :params new_variables_order_inner:
    :params old_variable_order:
    :params old_assign_probs: A dictionary of assignment and coresponding probabilities.
    Example:
    old_variable_order = [a, b]
    new_variables_order_outer = [b]

      a  b  c   P(a,b)     return:       b    a  c  P(b,a)
    {(0, 0, 0): pa0b0c0                {(0):{(0, 0): pa0b0,
     (0, 1, 0): pa0b1c0                      (1, 0): pa1b0}
     (1, 0, 1): pa1b0c1                 (1):{(0, 1): pa0b1,
     (1, 1, 1): pa1b1c1}                     (1, 1): pa1b1}}
    """
    # TODO: Complete and improve docstring.
    new_variable_order = new_variables_order_outer + new_variables_order_inner
    new_order_indices = [new_variable_order.index(var) for var in old_variable_order]
    new_assign_probs = dict()
    for old_assign_i, old_prob_i in old_assign_probs.items():
        new_row_assignment = [None] * len(old_assign_i)
        for old_i, new_i in enumerate(new_order_indices):
            new_row_assignment[new_i] = old_assign_i[old_i]
        l1_assign = tuple(new_row_assignment[:len(new_variables_order_outer)])
        if l1_assign not in new_assign_probs:
            new_assign_probs[l1_assign] = dict()
        assign_l2 = tuple(new_row_assignment[len(new_variables_order_outer):])
        new_assign_probs[l1_assign][assign_l2] = old_prob_i
    return new_assign_probs, new_variable_order


def same_scope_binary_operation(a, b, func, default):
    """
    NB: this function assumes that the variables corresponding to the keys in the two different dicts have the same order.
    :param a: The dictionary for factor (typically sub factor) A.
    :param b: The dictionary for factor (typically sub factor) B.
    """
    result_common_sub_dict = dict()
    all_common_assignments = set([*a.keys()] + [*b.keys()])
    new_default = func(default, default)
    for assignment in all_common_assignments:
        a_val = a.get(assignment, default)
        b_val = b.get(assignment, default)
        r_val = func(a_val, b_val)
        if r_val != new_default:
            result_common_sub_dict[assignment] = r_val
    return result_common_sub_dict


def flatten_ntd(ntd):
    """
    Flatten a two level nested table dictionary.
    """
    flattened_ntd = dict()
    for assign_outer in ntd.keys():
        for assign_inner in ntd[assign_outer].keys():
            flattened_ntd[assign_outer + assign_inner] = ntd[assign_outer][assign_inner]
    return flattened_ntd


def any_scope_binary_operation(ntd_a, outer_inner_cards_a,
                               ntd_b, outer_inner_cards_b,
                               func, default, default_rules='none'):
    """
    Apply a binary operation between categorical tables that have the same, disjoint or overlapping variable scopes.

    :param ntd_a: The nested table dictionary for factor A.
    :param ntd_b: The nested table dictionary for factor B.
    :param default: The default value for ntd_a and ntd_b (the values for the missing values).
    :param default_rules: The rules for when a calculation results will result in a default value (optional).
        This can help speed up this function. The possible values are as follows:
            'any' : If either ntd_a or ntd_b has a default value, the result will always be default.
            'both': Only if both ntd_a and ntd_b has a default value, the result will always be default.
            'none': No combination of default or non-default values is guarenteed to result in a default value
        If this parameter is not specified, 'none' will be used to ensure correct, albeit slower computation.
    :returns: The nested table dictionary for the resulting factor.
    """

    # TODO: this will not work for tables with the same scope, as this will mess up the nestedness(?).
    #  Seems like it works - confirm.

    resulting_factor_table = dict()
    ntd_a_sd_assignments = ntd_a.keys()
    ntd_b_sd_assignments = ntd_b.keys()
    # Note: sd: symmetric_difference

    full_sd_a_default_sub_table = None
    full_sd_b_default_sub_table = None

    # Calculate the symmetric_difference_combinations
    # The symmetric_difference_combinations are the combinations of the sub-assignments
    # that are not common to both factors. Depending on the value of default_rules, these
    # assignments could only include assignments actually specified in the nested table
    # dictionaries, or some or all of the missing assignments (if default values in either)
    # table does not necessarily guarantee default results.
    if default_rules == 'none':

        # we have to check everything - any combination of {value, default_value} could result in non-default value.
        ntd_a_sd_assignments = list(itertools.product(*[range(c) for c in outer_inner_cards_a[0]]))
        ntd_b_sd_assignments = list(itertools.product(*[range(c) for c in outer_inner_cards_b[0]]))

        full_sd_a_default_sub_table = _make_dense_default_probs_dict(outer_inner_cards_a[1], default)
        full_sd_b_default_sub_table = _make_dense_default_probs_dict(outer_inner_cards_b[1], default)

        # Having only dense default table is not good enough, as this will still not cause the missing
        # default values in the sub tables to be used to compute potentially none default result values.
        # we therefore need to make these sub tables dense.
        ntd_a = _make_inner_dense(ntd_a, outer_inner_cards_a, default)
        ntd_b = _make_inner_dense(ntd_b, outer_inner_cards_b, default)

    elif default_rules == 'both':
        # The result will only be default if the values in both ntd_a are default. This means
        # that we have to extend the assignments in the symmetric_difference_combinations to
        # those the full set of all possible assignments where either ntd_a is not default or
        # ntd_b is not default, because we need to calculate these results explicitly.
        # For example:
        # ntd_a:
        #    a b
        #    0 0  v
        #   (1 0) d
        # ntd_b:
        #    a c
        #    0 0  v
        #    1 0  v
        # The full set of resulting operations would be:
        # a b c
        # 0 0 0 vv
        # 0 0 1 vd
        # 0 1 0 dv
        # 0 1 1 dd = d
        # 1 0 0 dv
        # 1 0 1 dd = d
        # 1 1 0 dv
        # 1 1 1 dd = d
        # But because a:1 is not in ntd_a, this will not be calculated automatically and we therefore need to add this
        # entry. We do this by adding default sub tables below.
        outer_assignments_a = set(ntd_a.keys())
        outer_assignments_b = set(ntd_b.keys())
        ntd_a_sd_assignments = list(outer_assignments_a.union(outer_assignments_b))
        ntd_b_sd_assignments = copy.deepcopy(ntd_a_sd_assignments)

        full_sd_a_default_sub_table = _make_dense_default_probs_dict(outer_inner_cards_a[1], default)
        full_sd_b_default_sub_table = _make_dense_default_probs_dict(outer_inner_cards_b[1], default)

    # TODO: This is strange - improve.
    # TODO: See if any performance gains are possible with 'left' and 'right' flags and add them if so.
    elif default_rules == 'any':
        pass  # no further processing necessary
    symmetric_difference_combinations = itertools.product(*[ntd_a_sd_assignments, ntd_b_sd_assignments])

    for sd_assign_a, sd_assign_b in symmetric_difference_combinations:
        joined_sd_assignment = sd_assign_a + sd_assign_b

        # TODO: Add special, pre-computed result if both 0.
        common_vars_subtable_a = ntd_a.get(sd_assign_a, full_sd_a_default_sub_table)
        common_vars_subtable_b = ntd_b.get(sd_assign_b, full_sd_b_default_sub_table)

        # TODO: extend same_scope_binary_operation functionality to allow table completion given flag.
        subtable_result = same_scope_binary_operation(common_vars_subtable_a,
                                                      common_vars_subtable_b,
                                                      func, default)
        resulting_factor_table[joined_sd_assignment] = subtable_result
    return resulting_factor_table


def _fast_copy_probs_table(table):
    """
    Copy a dictionary representation of a probability table faster than the standard deepcopy.
    :param dict table: A dictionary with the tuples of ints as keys and floats as values.
    :return: The copied table.
    """
    table_copy = {tuple([a for a in assign]): value for assign, value in table.items()}
    return table_copy


class SparseCategorical(Factor):
    """
    A class for instantiating sparse tables with log probabilities.
    """

    def __init__(self, var_names, cardinalities, log_probs_table=None, probs_table=None, default_log_prob=-np.inf):
        """
        Construct a SparseLogTable. Either log_probs_table or probs_table should be supplied.

        :param var_names: The variable names.
        :type var_names: str list
        :param log_probs_table: A dictionary with assignments (tuples) as keys and probabilities as values.
            Missing assignments are assumed to have zero probability.
        :type
        :param log_probs_table: A dictionary with assignments (tuples) as keys and log probabilities as values.
            Missing assignments are assumed to have -inf log-probability (zero probability).
        Example:
            >>> var_names = ['rain','slip']
            >>> probs_table = {(0,0):0.8,
            >>>                (0,1):0.2,
            >>>                (1,0):0.4,
            >>>                (1,1):0.6}
            >>> cardinalities = [2,2]
            >>> table = SparseCategorical(log_probs_table=log_probs_table,
            >>>                           var_names=var_names,
            >>>                           cardinalities=cardinalities)
        """
        # TODO: add check that assignment lengths are consistent with var_names
        # TODO: add check that cardinalities are consistent with assignments
        super().__init__(var_names=var_names)
        if len(cardinalities) != len(var_names):
            raise ValueError('The cardinalities and var_names lists should be the same length.')
        if (log_probs_table is None) and (probs_table is None):
            raise ValueError('Either log_probs_table or probs_table must be specified')
        if log_probs_table is None:
            log_probs_table = {assignment: np.log(prob) for assignment, prob in probs_table.items()}
        self.log_probs_table = _fast_copy_probs_table(log_probs_table)
        self.var_cards = dict(zip(var_names, cardinalities))
        self.cardinalities = cardinalities
        self.default_log_prob = default_log_prob

    # TODO: Improve this to take missing assignments into account. Alternatively: add functionality to sparsify factor
    #  when probs turn to 0.
    # TODO: Add variable order sorting
    def equals(self, factor):
        """
        Check if this factor is the same as another factor.

        :param factor: The other factor to compare to.
        :type factor: SparseCategorical
        :return: The result of the comparison.
        :rtype: bool
        """
        factor_ = factor
        if not isinstance(factor_, SparseCategorical):
            raise ValueError(f'factor must be of SparseLogTable type but has type {type(factor)}')
            raise ValueError(f'factor must be of SparseLogTable type but has type {type(factor)}')

        if set(self.var_names) != set(factor_.var_names):
            return False

        # var sets are the same
        if self.var_names != factor.var_names:
            factor_ = factor.reorder(self.var_names)
        # factors now have same variable order

        # Check the values for every non-default assignment of self
        for assign, self_log_prob in self.log_probs_table.items():

            if assign not in factor_.log_probs_table:
                if self_log_prob != factor_.default_log_prob:
                    return False
            else:
                factor_log_prob = factor_.log_probs_table[assign]
                if not np.isclose(factor_log_prob, self_log_prob):
                    return False
        # everywhere that self has non default values, factor has the same values.

        # TODO: improve efficiency here (there could be a lot of duplication with the above loop)
        # Check the values for every non-default assignment of factor
        for assign, factor_log_prob in factor_.log_probs_table.items():

            if assign not in self.log_probs_table:
                if factor_log_prob != self.default_log_prob:
                    return False
            else:
                self_log_prob = self.log_probs_table[assign]
                if not np.isclose(self_log_prob, factor_log_prob):
                    return False

        # If all possible assignments have not been checked - check that the default values are the same
        if not self._is_dense() and not factor._is_dense():
            if self.default_log_prob != factor.default_log_prob:
                return False

        return True

    def copy(self):
        """
        Make a copy of this factor.
        :return: The copy of this factor.
        :rtype: SparseCategorical
        """
        return SparseCategorical(var_names=self.var_names.copy(),
                                 log_probs_table=_fast_copy_probs_table(self.log_probs_table),
                                 cardinalities=copy.deepcopy(self.cardinalities),
                                 default_log_prob=self.default_log_prob)

    @staticmethod
    def _get_shared_order_smaller_vars(smaller_vars, larger_vars):
        """
        larger_vars = ['a', 'c', 'd', 'b']
        smaller_vars = ['c', 'e', 'b']
        return ['c', 'b']
        """
        shared_vars = [v for v in smaller_vars if v in larger_vars]
        remaining_smaller_vars = list(set(larger_vars) - set(shared_vars))
        smaller_vars_new_order = shared_vars + remaining_smaller_vars
        return smaller_vars_new_order

    @staticmethod
    def _intersection_has_same_order(larger_vars, smaller_vars):
        """
        Check if the intersection of two lists has the same order in both lists.
        Will return true if either list is empty? SHOULD THIS BE THE CASE?
        """
        indices_of_smaller_in_larger = [larger_vars.index(v) for v in smaller_vars if v in larger_vars]
        if sorted(indices_of_smaller_in_larger) == indices_of_smaller_in_larger:
            return True
        return False

    # TODO: change back to log form
    def marginalize(self, vrs, keep=False):
        """
        Sum out variables from this factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: The resulting factor.
        :rtype: SparseCategorical
        """

        vars_to_keep = super().get_marginal_vars(vrs, keep)
        vars_to_sum_out = [v for v in self.var_names if v not in vars_to_keep]
        nested_table, nested_table_vars = _get_nested_sorted_probs(new_variables_order_outer=vars_to_keep,
                                                                   new_variables_order_inner=vars_to_sum_out,
                                                                   old_variable_order=self.var_names,
                                                                   old_assign_probs=self.log_probs_table)
        result_table = dict()
        for l1_assign, log_probs_table in nested_table.items():
            prob = special.logsumexp(list(log_probs_table.values()))
            result_table[l1_assign] = prob

        result_var_cards = copy.deepcopy(self.var_cards)
        for v in vars_to_sum_out:
            del result_var_cards[v]
        cardinalities = list(result_var_cards.values())
        return SparseCategorical(var_names=vars_to_keep, log_probs_table=result_table,
                                 cardinalities=cardinalities)

    def reduce(self, vrs, values):
        """
        Observe variables to have certain values and return reduced table.
        :param vrs: (list) The variables.
        :param values: (tuple or list) Their values
        :return: The resulting factor.
        :rtype: SparseCategorical
        """

        vars_unobserved = [v for v in self.var_names if v not in vrs]
        nested_table, nested_table_vars = _get_nested_sorted_probs(new_variables_order_outer=vrs,
                                                                   new_variables_order_inner=vars_unobserved,
                                                                   old_variable_order=self.var_names,
                                                                   old_assign_probs=self.log_probs_table)
        result_table = nested_table[tuple(values)]
        result_var_cards = copy.deepcopy(self.var_cards)
        for v in vrs:
            del result_var_cards[v]

        cardinalities = list(result_var_cards.values())
        return SparseCategorical(var_names=vars_unobserved, log_probs_table=result_table,
                                 cardinalities=cardinalities)

    def _assert_consistent_cardinalities(self, factor):
        """
        Assert that the variable cardinalities are consistent between two factors.

        :param factor:
        """
        for var in self.var_names:
            if var in factor.var_cards:
                error_msg = f'Error: inconsistent variable cardinalities: {factor.var_cards}, {self.var_cards}'
                assert self.var_cards[var] == factor.var_cards[var], error_msg

    def multiply(self, factor):
        """
        Multiply this factor with another factor and return the result.

        :param factor: The factor to multiply with.
        :type factor: SparseCategorical
        :return: The factor product.
        :rtype: SparseCategorical
        """
        return self._apply_binary_operator(factor, operator.add, default_rules='any')

    def cancel(self, factor):
        """
        Almost like divide, but with a special rule that ensures that division of zeros by zeros results in zeros.
        :param factor: The factor to divide by.
        :type factor: SparseCategorical
        :return: The factor quotient.
        :rtype: SparseCategorical
        """

        def special_divide(a, b):
            if (a == -np.inf) and (b == -np.inf):
                return -np.inf
            else:
                return a - b

        return self._apply_binary_operator(factor, special_divide, default_rules='any')

    def divide(self, factor):
        """
        Divide this factor by another factor and return the result.

        :param factor: The factor to divide by.
        :type factor: SparseCategorical

        :return: The factor quotient.
        :rtype: SparseCategorical
        """
        return self._apply_binary_operator(factor, operator.sub, default_rules='none')

    def argmax(self):
        """
        Get the Categorical assignment (vector value) that maximises the factor potential.

        :return: The argmax assignment.
        :rtype: int list
        """
        return max(self.log_probs_table.items(), key=operator.itemgetter(1))[0]

    def _apply_to_probs(self, func, include_assignment=False):
        for assign, prob in self.log_probs_table.items():
            if include_assignment:
                self.log_probs_table[assign] = func(prob, assign)
            else:
                self.log_probs_table[assign] = func(prob)

    def _apply_binary_operator(self, factor, operator_function, default_rules='none'):
        """
        Apply a binary operator function f(self.factor, factor) and return the result

        :param factor: The other factor to use in the binary operation.
        :type factor: SparseCategorical
        :param default_rules: The rules for when a calculation results will result in a default value (optional).
            This can help speed up this function. The possible values are as follows:
                'left' : If ntd_a has a default value, the result will always be default.
                'right' : If ntd_b has a default value, the result will always be default.
                'any' : If either ntd_a or ntd_b has a default value, the result will always be default.
                'both': Only if both ntd_a and ntd_b has a default value, the result will always be default.
                'none': No combination of default or non-default values is guarenteed to result in a default value
        If this parameter is not specified, 'none' will be used to ensure correct, albeit slower computation.
        :return: The resulting factor.
        :rtype: SparseCategorical
        """
        if not isinstance(factor, SparseCategorical):
            raise ValueError(f'factor must be of SparseLogTable type but has type {type(factor)}')
        self._assert_consistent_cardinalities(factor)
        intersection_vars = list(set(self.var_names).intersection(set(factor.var_names)))
        intersection_vars = sorted(intersection_vars)

        remaining_a_vars = list(set(self.var_names) - set(intersection_vars))
        ntd_a, _ = _get_nested_sorted_probs(new_variables_order_outer=remaining_a_vars,
                                            new_variables_order_inner=intersection_vars,
                                            old_variable_order=self.var_names,
                                            old_assign_probs=self.log_probs_table)

        remaining_b_vars = list(set(factor.var_names) - set(intersection_vars))
        ntd_b, _ = _get_nested_sorted_probs(new_variables_order_outer=remaining_b_vars,
                                            new_variables_order_inner=intersection_vars,
                                            old_variable_order=factor.var_names,
                                            old_assign_probs=factor.log_probs_table)

        # TODO: Add this functionality
        if self.default_log_prob != factor.default_log_prob:
            error_msg = 'Cases where self.default_value and factor.default_value differ are not yet supported.'
            raise NotImplementedError(error_msg)
        default_log_prob = self.default_log_prob

        def vars_to_cards(factor, var_names):
            return [factor.var_cards[v] for v in var_names]

        outer_inner_cards_a = [vars_to_cards(self, remaining_a_vars),
                               vars_to_cards(self, intersection_vars)]
        outer_inner_cards_b = [vars_to_cards(factor, remaining_b_vars),
                               vars_to_cards(factor, intersection_vars)]

        result_ntd = any_scope_binary_operation(ntd_a, outer_inner_cards_a,
                                                ntd_b, outer_inner_cards_b,
                                                operator_function, default_log_prob,
                                                default_rules=default_rules)
        flattened_result_table = flatten_ntd(result_ntd)
        result_var_names = remaining_a_vars + remaining_b_vars + intersection_vars
        result_var_cards = {**self.var_cards, **factor.var_cards}
        result_cardinalities = [result_var_cards[v] for v in result_var_names]
        result_default_log_prob = operator_function(default_log_prob, default_log_prob)
        return SparseCategorical(var_names=result_var_names, log_probs_table=flattened_result_table,
                                 cardinalities=result_cardinalities, default_log_prob=result_default_log_prob)

    def normalize(self):
        """
        Return a normalized copy of the factor.
        :return: The normalized factor.
        """

        factor_copy = self.copy()
        logz = special.logsumexp(list(factor_copy.log_probs_table.values()))
        for assign, prob in factor_copy.log_probs_table.items():
            factor_copy.log_probs_table[assign] = prob - logz
        return factor_copy

    def _is_dense(self):
        """
        Check if all factor assignments have non-default values.
        :return: The result of the check.
        :rtype: Bool
        """
        num_assignments = len(self.log_probs_table)
        max_possible_assignments = np.prod(self.cardinalities)
        if num_assignments == max_possible_assignments:
            return True
        return False

    @property
    def is_vacuous(self):
        """
        Check if this factor is vacuous (i.e uniform).
        :return: Whether the factor is vacuous or not.
        :rtype: bool
        """
        if self.distance_from_vacuous() < 1e-10:
            return True
        return False

    def kl_divergence(self, factor, normalize_factor=True):
        """
        Get the KL-divergence D_KL(P||Q) = D_KL(self||factor) between a normalized version of this factor and another factor.
        Reference https://infoscience.epfl.ch/record/174055/files/durrieuThiranKelly_kldiv_icassp2012_R1.pdf, page 1.

        :param factor: The other factor
        :type factor: Gaussian
        :param normalize_factor: Whether or not to normalize the other factor before computing the KL-divergence.
        :type normalize_factor: bool
        :return: The Kullback-Leibler divergence
        :rtype: float
        """

        def kld_from_log_elements(log_pi, log_qi):
            if log_pi == -np.inf:
                # lim_{p->0} p*log(p/q) = 0
                # lim_{p->0} p*log(p/q) = 0, with q = p
                # So if p_i = 0, kld_i = 0
                return 0.0
            else:
                kld_i = np.exp(log_pi) * (log_pi - log_qi)
                return kld_i

        log_P = self.normalize()
        log_Q = factor
        if normalize_factor:
            log_Q = factor.normalize()

        kld_factor = SparseCategorical._apply_binary_operator(log_P,
                                                              log_Q,
                                                              kld_from_log_elements,
                                                              default_rules='none')  # TODO: check this

        klds = list(kld_factor.log_probs_table.values())
        kld = np.sum(klds)

        if kld < 0.0:
            if np.isclose(kld, 0.0, atol=1e-5):
                #  this is fine (numerical error)
                return 0.0
            print('\nnormalize_factor = ', log_P)
            print('self = ')
            self.show()
            print('normalized_self = ')
            log_P.show()
            print('\nfactor = ')
            log_Q.show()
            raise ValueError(f'Negative KLD: {kld}')
        return kld

    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between this factor and a uniform copy of it.

        :return: The KL divergence.
        :rtype: float
        """
        # make uniform copy
        uniform_factor = self.copy()
        cards = list(uniform_factor.var_cards.values())
        uniform_log_prob = -np.log(np.product(cards))
        uniform_factor._apply_to_probs(lambda x: uniform_log_prob)
        kl = self.kl_divergence(uniform_factor, normalize_factor=False)
        if kl < 0.0:
            raise ValueError(f"kl ({kl}) < 0.0")
            self.show()
        return kl

    def potential(self, vrs, assignment):
        """
        Get the value of the factor for a specific assignment.

        :param assignment: The assignment
        :return: The value
        """
        assert set(vrs) == set(self.var_names), 'variables (vrs) do not match factor variables.'
        vrs_to_var_names_indices = [self.var_names.index(v) for v in vrs]
        var_names_order_assignments = tuple([assignment[i] for i in vrs_to_var_names_indices])
        return self.log_probs_table[var_names_order_assignments]

    def _to_df(self):
        log_probs_table = self.log_probs_table
        var_names = self.var_names
        df = pd.DataFrame.from_dict(log_probs_table.items()).rename(columns={0: 'assignment', 1: 'log_prob'})
        df[var_names] = pd.DataFrame(df['assignment'].to_list())
        df.drop(columns=['assignment'], inplace=True)
        return df

    def show(self, exp_log_probs=True):
        """
        Print the factor.

        :param exp_log_probs: Whether or no to exponentiate the log probabilities (to display probabilities instead of
        log-probabilities)
        :type exp_log_probs: bool
        """
        prob_string = 'log(prob)'
        if exp_log_probs:
            prob_string = 'prob'
        print(self.var_names, ' ', prob_string)
        for assignment, log_prob in self.log_probs_table.items():
            prob = log_prob
            if exp_log_probs:
                prob = np.exp(prob)
            print(assignment, ' ', prob)

    def reorder(self, new_var_names_order):
        """
        Reorder categorical table variables to a new order and reorder the associated probabilities
        accordingly.
        :param new_var_names_order: The new variable order.
        :type new_var_names_order: str list
        :return: The factor with new order.

        Example:
            old_variable_order = [a, b]
            new_variable_order = [b, a]

            a b P(a,b)  return    b a P(a,b)
            0 0  pa0b0            0 0  pa0b0
            0 1  pa0b1            0 1  pa1b0
            1 0  pa1b0            1 0  pa0b1
            1 1  pa1b1            1 1  pa1b1
        """

        new_order_indices = [self.var_names.index(var) for var in new_var_names_order]
        new_log_probs_table = dict()
        for assignment, value in self.log_probs_table.items():
            reordered_assignment = tuple(assignment[i] for i in new_order_indices)
            new_log_probs_table[reordered_assignment] = value
        reordered_cardinalities = [self.cardinalities[i] for i in new_order_indices]

        return SparseCategorical(var_names=new_var_names_order,
                                 log_probs_table=new_log_probs_table,
                                 cardinalities=reordered_cardinalities)


class SparseCategoricalTemplate(FactorTemplate):

    def __init__(self, log_probs_table, cardinalities, var_templates=None):
        """
        Create a Categorical factor template.

        :param log_probs_table: The log_probs_table that specifies the assignments and values for the template.
        :type log_probs_table: tuple:float dict
        :param var_templates: A list of formattable strings.
        :type var_templates: str list

        log_probs_table example:
        {(0, 0): 0.1,
         (0, 1): 0.3,
         (1, 0): 0.1,
         (1, 1): 0.5}
        """
        # TODO: Complete and improve docstring.
        super().__init__(var_templates=var_templates)
        self.log_probs_table = _fast_copy_probs_table(log_probs_table)
        self.cardinalities = cardinalities

    def make_factor(self, format_dict=None, var_names=None):
        """
        Make a factor with var_templates formatted by format_dict to create specific var names.

        :param format_dict: The dictionary to be used to format the var_templates strings.
        :type format_dict: str dict
        :return: The instantiated factor.
        :rtype: SparseCategorical
        """
        if (self._var_templates is None) and (var_names is None):
            raise ValueError(
                'var_names need to be supplied to make a factor from SparseCategoricalTemplate without var_templates')
        if format_dict is not None:
            assert var_names is None
            var_names = [vt.format(**format_dict) for vt in self._var_templates]
        cardinalities = list(self.cardinalities)
        factor = SparseCategorical(log_probs_table=self.log_probs_table,
                                   var_names=var_names, cardinalities=cardinalities)
        return factor



"""
A module for instantiating and performing operations on multivariate Gaussian and Gaussian mixture distributions.
"""
# System imports
import cmath
import copy
import operator

# Third-party imports
import numpy as np
import numdifftools as nd
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from scipy import special

# Local imports
from veroku.factors._factor import Factor
from veroku.factors import _factor_utils
from veroku.factors._factor_template import FactorTemplate


def assert_consistent_forms(gaussian):
    if gaussian.COVFORM:
        if gaussian.CANFORM:
            g_copy_covform = gaussian.copy()
            g_copy_covform.CANFORM = False
            g_copy_canform = gaussian.copy()
            g_copy_canform.COVFORM = False
            if not g_copy_covform.equals(g_copy_canform):
                g_copy_covform.CANFORM = True
                g_copy_canform.COVFORM = True
                print('\n\n\nFROM INCONSISTENCY:')

                print('\nCOVFORM FACTOR:')
                g_copy_covform._update_canform()
                g_copy_covform.show(update_covform=False, show_canform=True)
                print('\nCANFORM FACTOR:')
                g_copy_canform.show(update_covform=False, show_canform=True)


def make_random_gaussian(var_names, mean_range=[-10, 10], cov_range=[1, 10]):
    """
    Make a d dimensional random Gaussian by independently sampling the mean and covariance parameters from uniform
    distributions.

    :param var_names: The variable name of the factor
    :type var_names: str list
    :param mean_range: The range between which a mean will the uniformly sampled.
    :type mean_range: float list
    :param cov_range:The range between which a covariance will the uniformly sampled.
    :type cov_range: float list
    :return: The random Gaussian.
    :rtype: Gaussian
    """
    assert var_names, 'Error: var_names list cannot be empty.'
    dim = len(var_names)
    cov = np.random.uniform(cov_range[0], cov_range[1], [dim, dim]) * np.eye(dim, dim)
    mean = np.random.uniform(mean_range[0], mean_range[1], [dim, 1])
    random_gaussian = Gaussian(cov=cov, mean=mean, log_weight=0.0, var_names=var_names)
    return random_gaussian


def make_std_gaussian(var_names):
    """
    Make a d dimensional standard Gaussian.

    :param var_names: The variable name of the factor.
    :type var_names: str list
    :return: The standard Gaussian
    :rtype: Gaussian
    """
    assert var_names, 'Error: var_names list cannot be empty.'
    dim = len(var_names)
    cov = np.eye(dim, dim)
    mean = np.zeros([dim, 1])
    random_gaussian = Gaussian(cov=cov, mean=mean, log_weight=0.0, var_names=var_names)
    return random_gaussian


def split_gaussian(gaussian):
    """
    Split a Gaussian into a three component Gaussian Mixture, with different means.

    :param gaussian: The Gaussian distribution to split.
    :type gaussian: Gaussian
    :return: The split Gaussian Mixture.
    :rtype: GaussianMixture
    """
    if gaussian.dim != 1:
        raise NotImplementedError('Gaussian must be one dimensional.')
    weights = [1/3, 1/3, 1/3]
    full_cov = gaussian.get_cov()
    full_mean = gaussian.get_mean()

    covs = [w*full_cov for w in weights]
    means = [full_mean - np.sqrt(full_cov), full_mean, full_mean + np.sqrt(full_cov)]

    gaussians = []
    for mean, cov, weight in zip(means, covs, weights):
        gaussians.append(Gaussian(cov=cov,
                                  mean=mean,
                                  log_weight=np.log(weight),
                                  var_names=gaussian.var_names))
    gm = GaussianMixture(gaussians)
    return gm


def make_linear_gaussian(A, N, conditioning_var_templates, conditional_var_templates):
    """
    Make a linear Gaussian factor template.
    :param A: The linear transform matrix.
    :type A: Numpy array
    :param N: The additive noise covarannce matrix.
    :type A: Numpy array
    :param conditioning_var_templates: The list of formattable strings for the conditioning variables (i.e: ['var_a{i}_{t}', 'var_b{i}_{t}'])
    :param conditional_var_templates: The list of formattable strings for the conditional variables (i.e: ['var_c{i}_{t}', 'var_d{i}_{t}'])
    :return: The linear Gaussian Template
    :rtype: GaussianTemplate
    """

    # TODO: improve this by adding the correct equations for calculating the linear Gaussian parameters that do not rely
    #  on the dummy factor multiplication.
    X_dim = len(conditioning_var_templates)
    Sxx = np.eye(X_dim)
    Kxx = np.linalg.inv(Sxx)

    # LinGauss Investigation
    # A_1 = np.linalg.inv(A)
    # I = np.eye(2)
    # result = np.linalg.inv(Kxx@(I - np.linalg.inv(I + A_1@N@A.T@Kxx)))
    # print('result = ', result)

    ux = np.zeros([X_dim, 1])
    hx = Kxx @ ux
    uxTKxxux = ux.transpose() @ hx
    gx = - 0.5 * uxTKxxux + np.log(1.0) - 0.5 * np.log(np.linalg.det(2.0 * np.pi * Sxx))

    S = np.block([[Sxx, Sxx @ A.T],
                  [A @ Sxx, A @ Sxx @ A.T + N]])
    K_joint = np.linalg.inv(S)
    K_cpd = K_joint.copy()
    K_cpd[:X_dim, :X_dim] = K_cpd[:X_dim, :X_dim] - Kxx

    mean_joint = np.block([[ux], [A.T @ ux]])

    h_joint = K_joint @ mean_joint
    h_cpd = h_joint.copy()
    h_cpd[:X_dim] = h_cpd[:X_dim] - hx

    uTKux = mean_joint.T @ K_joint @ mean_joint
    g_joint = 0.5 * uTKux + np.log(1.0) - 0.5 * np.log(np.linalg.det(2.0 * np.pi * S))
    g_cpd = g_joint - gx
    var_names = conditioning_var_templates + conditional_var_templates

    return Gaussian(var_names=var_names, K=K_cpd, h=h_cpd, g=g_cpd)


def make_linear_gaussian_cpd_template(A, N, conditioning_var_templates, conditional_var_templates):
    assert len(conditioning_var_templates) == A.shape[0]
    assert N.shape[0] == A.shape[0]
    nlg = make_linear_gaussian(A, N, conditioning_var_templates, conditional_var_templates)
    var_templates = conditioning_var_templates + conditional_var_templates
    gaussian_parameters = {'K': nlg.get_K(), 'h': nlg.get_h(), 'g': 0}
    return GaussianTemplate(gaussian_parameters, var_templates)


############################################################################################################
#   Gaussian
############################################################################################################

class Gaussian(Factor):
    """
    A class for instantiating and performing operations om multivariate Gaussian distributions.
    """

    # pylint: disable=invalid-name
    def __init__(self, cov=None, mean=None, log_weight=None, K=None, h=None, g=None, var_names=None):
        """
        General constructor that can use either covariance form or canonical form parameters to construct a
        d-dimensional multivariate Gaussian object.

        :param cov: The covariance matrix (or variance scalar in 1-dimensional case).
        :type cov: dxd numpy.ndarray or float
        :param mean: The mean vector (or mean scalar in 1-dimensional case).
        :type mean: dx1 numpy.ndarray or float
        :param log_weight: the log of the weight (the Gaussian function integrates to the weight value)
        :type log_weight: float
        :param K: the precision matrix (or precision scalar in 1-dimensional case).
        :type K: dxd numpy.ndarray or float
        :param h: the h vector  (or h scalar in 1-dimensional case).
        :type h: dx1 numpy.ndarray or float
        :param g: the g parameter
        :type g: float
        :param var_names: a list of variable names where the order of the names correspond to the order in the
                         mean/covariance or K/h parameters
        :type var_names: str list
        """

        super().__init__(var_names=var_names)
        self._is_vacuous = False
        if all(v is None for v in [K, h, g]):
            if any(v is None for v in [cov, mean, log_weight]):
                raise ValueError('incomplete parameters')
            self.cov = _factor_utils.make_square_matrix(cov)
            self.mean = _factor_utils.make_column_vector(mean)
            assert len(self.cov.shape) == 2, "Error: Covariance matrix must be two dimensional."

            assert self.cov.shape[0] == self.dim, "Error: Covariance matrix dimension inconsistency."
            assert self.mean.shape[0] == self.dim, "Error: Mean vector dimension inconsistency."
            self.log_weight = _factor_utils.make_scalar(log_weight)
            self.K, self.h, self.g = None, None, None
            self.CANFORM = False
            self.COVFORM = True
        else:
            if any(v is None for v in [K, h, g]):
                raise ValueError('incomplete parameters')
            self.K = _factor_utils.make_square_matrix(K)
            self.h = _factor_utils.make_column_vector(h)
            if self.K.shape[0] != self.dim:
                pass
            assert len(self.K.shape) == 2, "Error: Precision matrix must be two dimensional."
            assert self.K.shape[0] == self.dim, "Error: Precision matrix dimension inconsistency."
            h_error_msg = f"Error: h vector dimension inconsistency: {self.h.shape[0]} (should be {self.dim})"
            assert self.h.shape[0] == self.dim, h_error_msg
            self.g = _factor_utils.make_scalar(g)
            self.cov, self.mean, self.log_weight = None, None, None
            self.CANFORM = True
            self.COVFORM = False

            # Note: this is important to allow vacuous Gaussians (that arise, for exmaple, when Gaussians are divided by
            # identical distributions or with unconditioned nonlinear Gaussians) to be merginalised.
            if self.is_vacuous:
                self._is_vacuous = True

    # pylint: enable=invalid-name

    @classmethod
    def make_vacuous(cls, var_names, g=0.0):
        """
        Make an vacuous Gaussian distribution with a zero precision matrix and zero h vector and zero g value.
        This 'Gaussian' is effectively a constant function (it has infinite variance and infinite mass) with value
        exp(g).

        :param var_names: The variable names.
        :type var_names: str list
        :param g: The g parameter.
        :type g: float
        :return: The vacuous Gaussian with zero K and h parameters.
        :rtype: Gaussian
        """
        dim = len(var_names)
        K_zero = np.zeros([dim, dim])
        h_zero = np.zeros([dim, 1])
        g_zero = g
        return cls(K=K_zero, h=h_zero, g=g_zero, var_names=var_names)

    # pylint: disable=invalid-name
    # pylint: disable=protected-access
    def _reorder_parameters(self, new_order_vars):
        """
        Reorder the values in the matrix and vector parameters according to the order of new_order_vars

        :param new_order_vars: The same variable names as in self.var_names, in a different order
        :type new_order_vars: str list
        """
        if new_order_vars == self._var_names:
            return
        assert set(new_order_vars) == set(self._var_names), \
            'Error: new_order_vars must contain the same variable names as in var_names.'
        new_order = [new_order_vars.index(var) for var in self._var_names]

        if self.CANFORM:
            new_K = np.zeros([self.dim, self.dim])
            new_h = np.zeros([self.dim, 1])
            for i, new_i in enumerate(new_order):
                new_h[new_i, 0] = self.h[i, 0]
                for j, new_j in enumerate(new_order):
                    new_K[new_i, new_j] = self.K[i, j]
            self.K = new_K
            self.h = new_h

        if self.COVFORM:
            new_cov = np.zeros([self.dim, self.dim])
            new_mean = np.zeros([self.dim, 1])
            for i, new_i in enumerate(new_order):
                new_mean[new_i, 0] = self.mean[i, 0]
                for j, new_j in enumerate(new_order):
                    new_cov[new_i, new_j] = self.cov[i, j]
            self.cov = new_cov
            self.mean = new_mean
        self._var_names = new_order_vars

    # pylint: enable=protected-access
    # pylint: enable=invalid-name

    # pylint: disable=protected-access
    def _canform_equals(self, gaussian, rtol, atol):
        """
        Helper function for check equivalence of canonical form parameters.
        """
        if not np.isclose(self.g, gaussian.get_g(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        gaussian._reorder_parameters(self.var_names)
        if not np.allclose(self.h, gaussian.get_h(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        if not np.allclose(self.K, gaussian.get_K(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        return True
    # pylint: enable=protected-access

    # pylint: disable=protected-access
    def _covform_equals(self, gaussian, rtol, atol):
        """
        Helper function for check equivalence of covariance form parameters.
        """
        if not np.isclose(self.log_weight, gaussian.get_log_weight(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        gaussian._reorder_parameters(self.var_names)
        if not np.allclose(self.mean, gaussian.get_mean(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        if not np.allclose(self.cov, gaussian.get_cov(), rtol=rtol, atol=atol, equal_nan=False):
            return False
        return True
    # pylint: enable=protected-access

    def equals(self, factor, rtol=1e-05, atol=1e-05):
        """
        Check if this factor is the same as another factor.

        :param factor: The factor to compare with.
        :type factor: Gaussian
        :param rtol: The absolute tolerance parameter (see numpy Notes for allclose function).
        :type rtol: float
        :param atol: The absolute tolerance parameter (see numpy Notes for allclose function).
        :type atol: float
        :return: Result of equals comparison between self and gaussian
        rtype: bool
        """
        # if not isinstance(factor, Gaussian):
        #    return False
        if set(self._var_names) != set(factor.var_names):
            return False

        gaussian_copy = factor.copy()
        # pylint: disable=protected-access
        if self._var_names != factor.var_names:
            gaussian_copy._reorder_parameters(self._var_names)
        if gaussian_copy._is_vacuous and self._is_vacuous:
            return True
        if self.CANFORM:
            if self._canform_equals(factor, rtol, atol):
                return True
            return False
        self._update_covform()
        if self.COVFORM:
            if not self._covform_equals(factor, rtol, atol):
                return False
        # pylint: enable=protected-access
        return True

    # pylint: disable=invalid-name
    def get_K(self):
        """
        Get the K parameter.

        :return: The K parameter.
        :rtype: dxd numpy.ndarray or float
        """
        self._update_canform()
        if self.K is not None:
            return self.K.copy()
        return None

    # pylint: enable=invalid-name

    # pylint: disable=invalid-name
    def get_h(self):
        """
        Get the h vector.

        :return: The h parameter.
        :rtype: float
        """
        self._update_canform()
        if self.h is not None:
            return self.h.copy()
        return None
    # pylint: enable=invalid-name

    # pylint: disable=invalid-name
    def get_g(self):
        """
        Get the g parameter.

        :return: The g parameter.
        :rtype: float
        """
        self._update_canform()
        return self.g
    # pylint: enable=invalid-name

    def cov_exists(self):
        if self.COVFORM:
            return True
        else:
            try:
                np.linalg.inv(self.K)
            except np.linalg.LinAlgError:
                return False
        return True

    # pylint: disable=invalid-name
    def get_cov(self):
        """
        Get the covariance parameter.

        :return: The cov parameter.
        :rtype: numpy.ndarray
        """
        self._update_covform()
        if self.cov is not None:
            return self.cov.copy()
        return None
    # pylint: enable=invalid-name

    def get_mean(self):
        """
        Get the mean parameter.

        :return: The mean parameter.
        :rtype: numpy.ndarray
        """
        self._update_covform()
        if self.mean is not None:
            return self.mean.copy()
        return None

    def get_log_weight(self):
        """
        Get the log weight parameter.
        :return: The log_weight parameter.
        :rtype: numpy.ndarray
        """
        self._update_covform()
        return self.log_weight

    def normalize(self):
        gaussian_copy = self.copy()
        gaussian_copy._update_covform()
        gaussian_copy.log_weight = 0.0
        gaussian_copy._destroy_canform()
        return gaussian_copy

    def get_complex_weight(self):
        """
        Get (possibly complex) weight parameter.

        :return: The weight parameter.
        :rtype: float
        """
        return cmath.exp(self._get_complex_log_weight())

    def _get_complex_log_weight(self):
        """
        Get the log weight even in cases where the determinant of the covariance matrix is negative. In such cases the
        log_weight no longer corresponds to the integral and the log of the weight will have a imaginary component.
        Computing the complex log-weight can however still be useful: it is used, for example, in the (experimental)
        Gaussian mixture division function (_gm_division_m2).
        """
        # pylint: disable=invalid-name
        self._update_canform()
        cov = np.linalg.inv(self.K)
        mean = cov.dot(self.h)
        uTKu = ((mean.transpose()).dot(self.K)).dot(mean).astype(complex)
        log_weight = self.g + 0.5 * uTKu + 0.5 * cmath.log(np.linalg.det(2.0 * np.pi * cov))
        return log_weight
        # pylint: enable=invalid-name

    def _invert(self):
        """
        Invert this Gaussian (1/Gaussian). This is used in the approximate Gaussian mixture division algorithms.

        :return: The inverted Gaussian.
        """
        assert self.CANFORM
        gaussian_copy = self.copy()
        gaussian_copy.K = (-1.0) * self.K
        gaussian_copy.h = (-1.0) * self.h
        gaussian_copy.g = (-1.0) * self.g
        gaussian_copy._destroy_covform()
        return gaussian_copy

    def _add_log_weight(self, log_weight_to_add):
        """
        Add log value to the log weight.

        :param log_weight_to_add: The log value to add to the weight.
        :type log_weight_to_add: float
        """
        if self.CANFORM:
            self.g += log_weight_to_add
        if self.COVFORM:
            self.log_weight += log_weight_to_add

    def get_weight(self):
        """
        Get the weight of the distribution - the value of the integral of the (possibly unnormalized) distribution.

        :return: The weight.
        :rtype: float
        """
        return np.exp(self.get_log_weight())

    def _fix_non_psd_matrices(self):

        # TODO: how to ensure these are consistent?
        if self.COVFORM:
            if not _factor_utils.is_pos_def(self.cov):
                self.cov = _factor_utils.get_closest_psd_from_symmetric_matrix(self.cov)

        if self.CANFORM:
            if not _factor_utils.is_pos_def(self.K):
                self.K = _factor_utils.get_closest_psd_from_symmetric_matrix(self.K)

    def marginalize(self, vrs, keep=False):
        """
        Integrate out variables from this Gaussian.

        :param vrs: A subset of variables in the factor's scope.
        :type vrs: str list
        :param keep: Whether to keep or sum out vrs.
        :type keep: bool
        :return: The resulting Gaussian marginal.
        :rtype: Gaussian
        """
        vars_to_keep = super().get_marginal_vars(vrs, keep)
        vars_to_integrate_out = list(set(self.var_names) - set(vars_to_keep))
        if self._is_vacuous:
            # TODO: check this (esp log_weight)
            # print('Warning: marginalising vacuous Gaussian')
            return Gaussian.make_vacuous(var_names=vars_to_keep)

        if self.CANFORM:
            xx_indcs = [self.var_names.index(var_x) for var_x in vars_to_keep]
            yy_indcs = [self.var_names.index(var_y) for var_y in vars_to_integrate_out]
            Kxx = self.K[np.ix_(xx_indcs, xx_indcs)]
            Kxy = self.K[np.ix_(xx_indcs, yy_indcs)]
            Kyx = Kxy.transpose()
            Kyy = self.K[yy_indcs][:, yy_indcs]
            try:
                inv_Kyy = np.linalg.inv(Kyy)
            except np.linalg.LinAlgError as e:
                print('vars_to_integrate_out = ', vars_to_integrate_out)
                self.show_vis(figsize=(20, 16))
                # return Gaussian.make_vacuous(var_names=vars_to_keep)
                raise e
            #    # TODO: check this (not sure if it is correct)
            #    return self.make_vacuous(vars_to_keep, g=self.g)
            hx = self.h[xx_indcs]
            hy = self.h[yy_indcs]
            Kxy_inv_Kyy = Kxy.dot(np.linalg.inv(Kyy))
            K = Kxx - Kxy_inv_Kyy.dot(Kyx)
            h = hx - Kxy_inv_Kyy.dot(hy)
            g = self.g + 0.5 * ((hy.T.dot(np.linalg.inv(Kyy)).dot(hy)) + np.log(np.linalg.det(2.0 * np.pi * inv_Kyy)))
            return Gaussian(K=K, h=h, g=g, var_names=vars_to_keep)

        self._update_covform()
        indices_to_keep = [self._var_names.index(variable) for variable in vars_to_keep]
        marginal_var_names = vars_to_keep.copy()

        if self._is_vacuous:
            raise ValueError('cannot marginalize vacuous Gaussian.')

        marginal_cov = self.cov[np.ix_(indices_to_keep, indices_to_keep)]
        marginal_mean = self.mean[np.ix_(indices_to_keep, [0])]
        return Gaussian(cov=marginal_cov, mean=marginal_mean, log_weight=self.log_weight,
                        var_names=marginal_var_names)

    def _destroy_canform(self):
        self.K, self.h, self.h = None, None, None
        self.CANFORM = False

    def _destroy_covform(self):
        self.cov, self.mean, self.log_weight = None, None, None
        self.COVFORM = False

    def _update_canform(self):
        """
        Update the canonical form parameters of the Gaussian.
        """
        # pylint: disable=invalid-name
        if self.CANFORM:
            return
        assert self.COVFORM
        try:
            self.K = np.linalg.inv(self.cov)
        except np.linalg.LinAlgError as e:
            print('self.cov = ', self.cov)
            raise e
        self.h = self.K.dot(self.mean)
        uTKu = ((self.mean.transpose()).dot(self.K)).dot(self.mean)

        try:
            det2picov = np.linalg.det(2.0 * np.pi * self.cov)
            if det2picov < 0.0:
                raise ValueError(f'invalid value in log: det2picov={det2picov} ')
            g_ = self.log_weight - 0.5 * uTKu - 0.5 * np.log(det2picov)
        except Exception as e:
            print(self.cov)
            raise e
        self.g = _factor_utils.make_scalar(g_)
        self.CANFORM = True
        # pylint: enable=invalid-name

    def _update_covform(self):
        """
        Update the covariance form parameters of the Gaussian.
        """
        if self._is_vacuous:
            raise Exception('cannot update covariance form for vacuous Gaussian.')
        # pylint: disable=invalid-name
        if self.COVFORM:
            return
        assert self.CANFORM
        try:
            self.cov = np.linalg.inv(self.K)
        except np.linalg.LinAlgError as e:
            np.set_printoptions(edgeitems=20, precision=2, linewidth=500)
            print('cannot update cov form: self.K \n= ', self.var_names, '\n', self.K)
            # self.show_vis()
            print(e)
            raise e
        assert not np.isnan(np.sum(self.cov)), 'Error: nan values in cov matrix after inversion.'
        self.mean = self.cov.dot(self.h)
        uTKu = ((self.mean.transpose()).dot(self.K)).dot(self.mean)
        det_2_pi_cov = np.linalg.det(2.0 * np.pi * self.cov)
        # if det_2_pi_cov < 0:
        #    print('Error: 0 > det_2_pi_cov = ', det_2_pi_cov)
        log_weight_ = self.g + 0.5 * uTKu + 0.5 * np.log(det_2_pi_cov)
        self.log_weight = _factor_utils.make_scalar(log_weight_)
        self.COVFORM = True
        # pylint: enable=invalid-name

    def multiply(self, factor):
        """
        Multiply this Gaussian with another factor.
        :param factor: the factor to multiply with
        :return: the resulting factor
        """
        # if isinstance(factor, NonLinearGaussian):
        #    return factor.multiply(self)
        return self._absorb_or_cancel(factor, operator.add)

    def divide(self, factor):
        """
        Divide this Gaussian by another facotr.
        :param factor: the factor to divide by
        :return: the resulting factor
        """
        return self._absorb_or_cancel(factor, operator.sub)

    def _absorb_or_cancel(self, factor, operator_function):
        """
        A general function which can either perform Gaussian multiplication or division (which are very similar).
        :param factor: the gaussian to multiply or divide by
        :param operator_function: the operator to use one the Gaussian canonical parameters
                        ('+' for multiplication and '-' for division)
        :return: the resulting GaussianMixture
        """
        # pylint: disable=invalid-name
        K_a = self.get_K()
        K_b = factor.get_K()
        assert len(K_a.shape) == 2
        assert len(K_b.shape) == 2
        K_c, new_vars_0 = _factor_utils.indexed_square_matrix_operation(K_a, K_b, self._var_names,
                                                                       factor.var_names, operator_function)
        g_a = self.get_g()
        g_b = factor.get_g()
        g_c = operator_function(g_a, g_b)

        h_a = self.get_h()
        h_b = factor.get_h()
        h_c, new_vars_1 = _factor_utils.indexed_column_vector_operation(h_a, h_b, self._var_names,
                                                                       factor.var_names, operator_function)
        assert new_vars_0 == new_vars_1
        return Gaussian(K=K_c, h=h_c, g=g_c, var_names=new_vars_0)
        # pylint: enable=invalid-name

    def sample(self, num_samples):
        """
        Draw samples from the Gaussian distribution.
        :param num_samples: The number of samples to draw.
        :type num_samples:
        :return: The samples.
        :rtype: int
        """
        std_gaussian_samples = np.random.normal(0, 1, [self.dim, num_samples])
        Xs = std_gaussian_samples
        self._update_covform()
        L = np.linalg.cholesky(self.cov)
        X = L.dot(Xs) + self.mean
        return X

    # pylint: disable=invalid-name
    def _get_observation_reduced_canonical_vars(self, observed_indices, unobserved_indices, observed_vec):
        """
        A helper function for reduce.
        :param observed_indices: The observed variable indices
        :type observed_indices: int list
        :param unobserved_indices: The unobserved variable indices
        :type unobserved_indices: int list
        :return: the reduced parameters K_observed, h_observed, g_observed
        :rtype: numpy.ndarray, numpy.ndarray, float
        """
        K = self.get_K()
        h = self.get_h()
        K_reduced = 0
        h_reduced = 0
        if unobserved_indices:
            K_XX = K[np.ix_(unobserved_indices, unobserved_indices)]
            K_reduced = K_XX.copy()
            K_XY = K[np.ix_(unobserved_indices, observed_indices)]
            h_X = h[np.ix_(unobserved_indices, [0])]
            h_reduced = h_X - K_XY.dot(observed_vec)

        K_YY = K[np.ix_(observed_indices, observed_indices)]
        h_Y = h[np.ix_(observed_indices, [0])]
        g_reduced = self.get_g() + (h_Y.transpose()).dot(observed_vec) - 0.5 * (
            (observed_vec.transpose()).dot(K_YY)).dot(observed_vec)

        return K_reduced, h_reduced, g_reduced
    # pylint: enable=invalid-name

    # pylint: disable=invalid-name
    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this Gaussian and return the resulting factor.

        :param vrs: the names of the observed variable (list)
        :type vrs: str list
        :param values: the values of the observed variables
        :type values: vector-like
        :return: the resulting Gaussian
        :rtype: Gaussian
        """
        observed_vec = _factor_utils.make_column_vector(values)

        assert isinstance(vrs, list)  # just to future-proof interface
        assert set(vrs) <= set(self._var_names), \
            'Error: observed variables must be a subset of the gaussian variables.'

        unobserved_vars = list(set(self._var_names) - set(vrs))
        unobserved_vars.sort()  # the above operations seems to return inconsistent orderings
        observed_indices = [self._var_names.index(v) for v in vrs]
        unobserved_indices = [self._var_names.index(v) for v in unobserved_vars]

        K_reduced, h_reduced, g_reduced = self._get_observation_reduced_canonical_vars(
            observed_indices=observed_indices,
            unobserved_indices=unobserved_indices,
            observed_vec=observed_vec)

        return Gaussian(K=K_reduced, h=h_reduced, g=g_reduced, var_names=unobserved_vars)
        # pylint: enable=invalid-name

    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between this factor and a uniform copy of it.
        Note: here it does not matter if we take KL(P||Q) or KL(Q||P) the result is either 0.0 (if both are vacuous)
        or inf (if one is not).
        :return: The KL divergence.
        :rtype: float
        """
        if self._is_vacuous:
            return 0.0
        else:
            return float('inf')

    def kl_divergence(self, factor, normalize_factor=True):
        """
        Get the KL-divergence D_KL(self||factor) between a normalized version of this factor and another factor.
        Reference https://infoscience.epfl.ch/record/174055/files/durrieuThiranKelly_kldiv_icassp2012_R1.pdf, page 1.

        :param factor: The other factor
        :type factor: Gaussian
        :param normalize_factor: Whether or not to normalize the other factor before computing the KL-divergence.
        :type normalize_factor: bool
        :return: The Kullback-Leibler divergence
        :rtype: float
        """
        if self.dim != factor.dim:
            raise ValueError('cannot get KL-divergence between Gaussians of different dimensionalities.')
        if self._is_vacuous and factor._is_vacuous:
            return 0.0
        if self._is_vacuous or factor._is_vacuous:
            return np.inf

        if self.equals(factor):
            return 0.0
        # TODO: can we compute the correct ('normalised') KL divergence without explicitly normalizing?
        normalized_self = self.normalize()
        factor_ = factor
        if normalize_factor:
            factor_ = factor.normalize()
        inv_cov_q = factor_.get_K()
        inv_cov_p = normalized_self.get_K()
        cov_p = normalized_self.get_cov()

        u_q = factor_.get_mean()
        u_p = normalized_self.get_mean()

        det_inv_cov_q = np.linalg.det(inv_cov_q)
        det_inv_cov_p = np.linalg.det(inv_cov_p)
        if det_inv_cov_q == 0.0:
            if det_inv_cov_p != 0.0:
                return np.inf
            else:
                det_term = 0.0
        else:
            det_term = 0.5 * cmath.log(det_inv_cov_p / det_inv_cov_q)
        trace_term = 0.5 * np.trace(inv_cov_q.dot(cov_p))
        mahalanobis_term = 0.5 * (u_p - u_q).T.dot(inv_cov_q).dot(u_p - u_q)
        dim_term = 0.5 * normalized_self.dim
        kld = det_term + trace_term + mahalanobis_term - dim_term
        # TODO: Add warning or error if this is negative and remove abs below
        return np.abs(kld[0][0])

    @property
    def is_vacuous(self):
        """
        Check if a Gaussian distribution contains no information. This is the case when the K matrix is a zero matrix.
        :return: The result of the check.
        :rtype: Bool
        """

        if self._is_vacuous:
            return True
        if self.CANFORM:
            if np.allclose(self.K, 0.0):
                if not _factor_utils.is_pos_def(self.K):
                    return True
        return False

    def copy(self):
        """
        Make a copy of this Gaussian.

        :return: The copied Gaussian.
        :rtype: Gaussian
        """

        if self.COVFORM and self.CANFORM:
            assert isinstance(self.g, (int, float))
            assert isinstance(self.log_weight, (int, float))
            gaussian_copy = Gaussian(cov=self.cov.copy(),
                                     mean=self.mean.copy(),
                                     log_weight=self.log_weight,
                                     var_names=copy.deepcopy(self._var_names))
            gaussian_copy.K = self.K.copy()
            gaussian_copy.h = self.h.copy()
            gaussian_copy.g = self.g
            gaussian_copy.CANFORM = True
            return gaussian_copy

        if self.COVFORM:
            assert isinstance(self.log_weight, (int, float))
            return Gaussian(cov=self.cov.copy(),
                            mean=self.mean.copy(),
                            log_weight=self.log_weight,
                            var_names=copy.deepcopy(self._var_names))
        if self.CANFORM:
            assert isinstance(self.g, (int, float))
            return Gaussian(K=self.K.copy(),
                            h=self.h.copy(),
                            g=self.g,
                            var_names=copy.deepcopy(self._var_names))
        raise Exception('Gaussian is neither in canonical form nor in covariance form?')

    def potential(self, x_val):
        """
        Get the value of the Gaussian potential at x_val.

        :param x_val: The vector (or vector-like object) to evaluate the Gaussian at
        :type x_val: numpy.ndarray
        :return: The value of the Gaussian potential at x_val.
        """
        return np.exp(self.log_potential(x_val))

    # pylint: disable=invalid-name
    def log_potential(self, x_val, vrs=None):
        """
        Get the log of the value of the Gaussian potential at x_val.

        :param x_val: the vector (or vector-like object) to evaluate the GaussianMixture at
        :type x_val: vector-like
        :param vrs: The variables corresponding to the values in x_val.
        :type vrs: str list
        :return: The log of the value of the GaussianMixture potential at x_val.
        """
        if vrs is not None:
            assert set(vrs) == set(self.var_names)
            if isinstance(x_val, np.ndarray):
                x_val = x_val.ravel()
            x_val_list = list(x_val)
            assert len(x_val_list) == len(self.var_names)
            x_val = [x_val_list[vrs.index(v)] for v in self.var_names]
        x_vec = _factor_utils.make_column_vector(x_val)
        if self.COVFORM:
            log_norm = self.log_weight - 0.5 * np.log(np.linalg.det(2.0 * np.pi * self.cov))
            K = np.linalg.inv(self.cov)
            exponent = ((-0.5 * (self.mean - x_vec).transpose()).dot(K)).dot(self.mean - x_vec)
            log_potx = log_norm + exponent

        if self.CANFORM:
            log_potx = -0.5 * (x_vec.transpose().dot(self.K)).dot(x_vec) + x_vec.transpose().dot(self.h) + self.g

        return log_potx[0, 0]

    # pylint: enable=invalid-name

    def get_cov_repr_str(self):
        self_copy = self.copy()
        self_copy._update_covform()
        np.set_printoptions(linewidth=np.inf)
        repr_str = 'Cov        = \n' + str(self_copy.cov) + '\n' + \
                   'mean       = \n' + str(self_copy.mean) + '\n' + \
                   'log_weight = \n' + str(self_copy.log_weight) + '\n'
        return repr_str

    def get_can_repr_str(self):
        self_copy = self.copy()
        self_copy._update_canform()
        np.set_printoptions(linewidth=np.inf)
        repr_str = 'K = \n' + str(self_copy.K) + '\n' + \
                   'h = \n' + str(self_copy.h) + '\n' + \
                   'g = \n' + str(self_copy.g) + '\n' + \
                   'is_vacuous: ' + str(self_copy._is_vacuous) + '\n'
        return repr_str

    def __repr__(self):  # pragma: no cover
        """
        Get the string representation of the Gaussian factor.
        """
        np.set_printoptions(edgeitems=3)
        np.set_printoptions(precision=4)
        np.core.arrayprint._line_width = 200
        repr_str = 'vars = ' + str(self.var_names) + '\n'
        if not self._is_vacuous:
            repr_str += self.get_can_repr_str()
        repr_str += self.get_cov_repr_str()
        return repr_str

    def show(self, update_covform=True, show_canform=False):  # pragma: no cover
        """
        Print the parameters of the Gaussian distribution

        :param update_covform: Whether or not to update the covariance form before showing.
        :type update_covform: bool
        :param show_canform: Whether or not toshow the canonical form as well.
        :type show_canform: bool
        """
        np.set_printoptions(edgeitems=3)
        np.set_printoptions(precision=4)
        np.core.arrayprint._line_width = 200
        self_copy = self.copy()
        if not self._is_vacuous and update_covform:
            self_copy._update_covform()
        print('vars = ', self_copy.var_names)
        if self_copy.COVFORM:
            print(self_copy.get_cov_repr_str())
        if self_copy.CANFORM and show_canform:
            print(self_copy.get_can_repr_str())

    def show_vis(self, figsize=(10, 8)):
        """
        Visualise the parameters with plots.

        :param figsize: The figure size.
        :type figsize: 2 element int tuple
        """
        # TODO: find a way of making cov matrix square.
        f, [ax_cov, ax_mean] = plt.subplots(nrows=2, figsize=figsize)  # ,sharex=True)
        cov_df = pd.DataFrame(data=self.get_cov(), index=self.var_names, columns=self.var_names)
        mean_df = pd.DataFrame(data=self.get_mean(), index=self.var_names, columns=['var_names'])
        sns.heatmap(cov_df, ax=ax_cov, cbar=True, cbar_kws=dict(use_gridspec=False, location="top"), annot=True)

        mean_df.plot.bar(ax=ax_mean, legend=False)
        plt.xticks(rotation=0)

    def _get_limits_for_2d_plot(self):  # pragma: no cover
        """
        Get x and y limits which includes most of the Gaussian mixture's mass, by considering
        the mean and variance of each Gaussian component.
        """
        self_copy = self.copy()
        self_copy._update_covform()
        x_lower = self_copy.mean[0, 0] - 3.0 * np.sqrt(self_copy.cov[0, 0])
        x_upper = self_copy.mean[0, 0] + 3.0 * np.sqrt(self_copy.cov[0, 0])
        y_lower = self_copy.mean[1, 0] - 3.0 * np.sqrt(self_copy.cov[1, 1])
        y_upper = self_copy.mean[1, 0] + 3.0 * np.sqrt(self_copy.cov[1, 1])
        return [x_lower, x_upper], [y_lower, y_upper]

    # TODO: reconcile with GaussianMixture _plot_2d
    def _plot_2d(self, log, xlim, ylim):  # pragma: no cover
        """
        Plot a 2d Gaussian mixture potential function

        :param log: if this is True, the log-potential will be plotted
        :param xlim: the x limits to plot the function over (optional)
        :param ylim: the y limits to plot the function over (optional)
        """
        if xlim is None and ylim is None:
            xlim, ylim = self._get_limits_for_2d_plot()
        elif xlim is not None and ylim is not None:
            pass
        else:
            print('Warning: partial limits received. Generating limits automatically.')
            xlim, ylim = self._get_limits_for_2d_plot()

        xlabel = self.var_names[0]
        ylabel = self.var_names[1]

        if not log:
            _factor_utils.plot_2d(func=self.potential, xlim=xlim, ylim=ylim, xlabel=xlabel, ylabel=ylabel)
        else:
            _factor_utils.plot_2d(func=self.log_potential, xlim=xlim, ylim=ylim, xlabel=xlabel, ylabel=ylabel)

    def plot(self, log=False, xlim=None, ylim=None):
        """
        Plot the Gaussian mixture potential function (only for 1d and 2d functions)

        :param log: if this is True, the log-potential will be plotted
        :param xlim: the x limits to plot the function over (optional)
        :type xlim: 2 element float list
        :param ylim: the y limits to plot the function over (optional and only used in 2d case)
        :type ylim: 2 element float list
        """
        # TODO: Replace with gaussian mixture plot.
        if self.dim == 1:
            if xlim is None:
                stddev = np.sqrt(self.get_cov()[0, 0])
                lb = self.get_mean()[0, 0] - 3 * stddev
                ub = self.get_mean()[0, 0] + 3 * stddev
                xlim = [lb, ub]
            if xlim is not None:
                x_lower = xlim[0]
                x_upper = xlim[1]
            num_points = 200
            x_series = np.linspace(x_lower, x_upper, num_points)
            if log:
                potx = np.array([self.log_potential(xi) for xi in x_series])
            else:
                potx = np.array([self.potential(xi) for xi in x_series])
            plt.plot(x_series, potx)
        elif self.dim == 2:
            self._plot_2d(log=log, xlim=xlim, ylim=ylim)
        else:
            raise NotImplementedError('Plotting not implemented for dim!=1.')


class GaussianTemplate(FactorTemplate):

    def __init__(self, gaussian_parameters, var_templates):
        """
        Create a Categorical factor template.

        :param log_probs_table: The log_probs_table that specifies the assignments and values for the template.
        :type log_probs_table: tuple:float dict
        :param var_templates: A list of formattable strings.
        :type var_templates: str list

        >>>gaussian_parameters = {'K': np.array([1,0][0,1]]),
        >>>                       'h': np.array([0][1]]),
        >>>                       'g': 0}
        """
        # TODO: Complete and improve docstring.
        super().__init__(var_templates=var_templates)
        self.K = gaussian_parameters['K']
        self.h = gaussian_parameters['h']
        self.g = gaussian_parameters['g']

    def make_factor(self, format_dict=None, var_names=None):
        """
        Make a factor with var_templates formatted by format_dict to create specific var names.

        :param format_dict: The dictionary to be used to format the var_templates strings.
        :type format_dict: str dict
        :return: The instantiated factor.
        :rtype: Categorical
        """
        if format_dict is not None:
            assert var_names is None
            var_names = [vt.format(**format_dict) for vt in self._var_templates]
        # TODO: remove this and find better solution
        g = Gaussian(K=self.K.copy(), h=self.h.copy(), g=self.g, var_names=var_names)
        g._is_vacuous = True
        return g


############################################################################################################
#   Gaussian Mixture
############################################################################################################

class GaussianMixture(Factor):
    """
    A Class for instantiating and performing operations on multivariate Gaussian mixture functions.
    """

    def __init__(self, factors, cancel_method=0):
        """
        A GaussianMixture constructor.

        :param factors: a list of Gaussian type objects with the same dimensionality.
        :type factors: Gaussian list
        :param cancel_method: The method of performing approximate division.
                              This is only applicable if the factor is a Gaussian mixture with more than one component.
                              0: moment match denominator to single Gaussian and divide
                              1: approximate modes of quotient as Gaussians
                              2: Use complex moment matching on inverse mixture sets (s_im) and approximate the inverse
                                 of each s_im as a Gaussian
        :type cancel_method: int
        """
        assert factors, 'Error: empty list passed to constructor.'
        self.cancel_method = cancel_method
        self.components = [gaussian.copy() for gaussian in factors]
        self.num_components = len(factors)

        var_names0 = factors[0].var_names

        for component in self.components:
            if var_names0 != component.var_names:
                raise ValueError('inconsistent var_names in list of Gaussians.')
        super().__init__(var_names=var_names0)

    def equals(self, factor):
        """
        Check if this Gaussian mixture is the same as another factor.

        :param factor: The factor to compare to.
        :type factor: GaussianMixture
        :return: The result of the comparison.
        :rtype: bool
        """
        if not isinstance(factor, GaussianMixture):
            return False
        if factor.num_components != self.num_components:
            return False

        for i in range(self.num_components):
            found_corresponding_factor = False
            for j in range(i, self.num_components):
                if self.get_component(i).equals(factor.get_component(j)):
                    found_corresponding_factor = True
            if not found_corresponding_factor:
                return False
        return True

    def get_component(self, index):
        """
        Get the Gaussian component at an index.

        :param index: The index of teh component to return.
        :type index: int
        :return: The component at the given index.
        :rtype: Gaussian
        """
        return self.components[index]

    def copy(self):
        """
        Make a copy of this Gaussian mixture.

        :return: The copied GaussianMixture.
        :rtype: GaussianMixture
        """
        component_copies = []
        for gauss in self.components:
            component_copies.append(gauss.copy())
        return GaussianMixture(component_copies)

    def multiply(self, factor):
        """
        Multiply this GaussianMixture with another factor.

        :param factor: The factor to multiply with.
        :type factor: Gaussian or Gaussian Mixture
        :return: The factor product.
        :rtype: GaussianMixture
        """
        new_componnents = []
        if isinstance(factor, Gaussian):
            for gauss in self.components:
                new_componnents.append(gauss.multiply(factor))
        elif isinstance(factor, GaussianMixture):
            for gauss_ai in self.components:
                for gauss_bi in factor.components:
                    new_componnents.append(gauss_ai.multiply(gauss_bi))
        else:
            raise TypeError('unsupported factor type.')
        return GaussianMixture(new_componnents)

    def divide(self, factor):
        """
        Divide this GaussianMixture by another factor.

        :param factor: The factor divide by.
        :type factor: Gaussian or Gaussian Mixture
        :return: The resulting factor quotient (approximate in the case of where both the numerator and denominator are
        GaussianMixtures with more than one component).
        :rtype: GaussianMixture
        """
        if isinstance(factor, Gaussian):
            single_gaussian = factor
        elif isinstance(factor, GaussianMixture):
            if factor.num_components == 1:
                single_gaussian = factor.get_component(index=0)
            else:
                if self.cancel_method == 0:
                    single_gaussian = factor.moment_match_to_single_gaussian()
                if self.cancel_method == 1:
                    return GaussianMixture._gm_division_m1(self, factor)
                if self.cancel_method == 2:
                    return GaussianMixture._gm_division_m2(self, factor)
        else:
            raise TypeError('unsupported factor type.')
        new_components = []
        for gauss in self.components:
            new_components.append(gauss.divide(single_gaussian))
        return GaussianMixture(new_components)

    def reduce(self, vrs, values):
        """
        Observe a subset of the variables in the scope of this Gaussian mixture and return the resulting factor.

        :param vrs: the names of the observed variable (list)
        :type vrs: str list
        :param values: the values of the observed variables (list or vector-like object)
        :type values: vactor-like
        :return: the observation reduced factor.
        :rtype: GaussianMixture
        """
        new_componnents = []
        for gauss in self.components:
            new_componnents.append(gauss.reduce(vrs, values))
        return GaussianMixture(new_componnents)

    def marginalize(self, vrs, keep=False):
        """
        Integrate out variables from this Gaussian mixture.

        :param vrs: A subset of variables in the factor's scope.
        :type vrs: str list
        :param keep: Whether to keep or sum out vrs.
        :type keep: bool
        :return: the resulting marginal factor.
        :rtype: GaussianMixture
        """
        new_componnents = []
        for gauss in self.components:
            new_componnents.append(gauss.marginalize(vrs, keep))
        return GaussianMixture(new_componnents)

    # pylint: disable=invalid-name
    def log_potential(self, x):
        """
        Get the log of the value of the Gaussian mixture potential at X.

        :param x: The point to evaluate the GaussianMixture at.
        :type x: vector-like
        :return: log of the value of the GaussianMixture potential at x.
        :rtype: float
        """
        log_potentials = []
        for gauss in self.components:
            log_potentials.append(gauss.log_potential(x))
        total_log_potx = special.logsumexp(log_potentials)
        return total_log_potx
    # pylint: enable=invalid-name

    def potential(self, x_val):
        """
        Get the value of the Gaussian mixture potential at x_val.

        :param x_val: The point to evaluate the GaussianMixture at.
        :type x_val: vector-like
        :return: log of the value of the GaussianMixture potential at x_val.
        :rtype: float
        """
        total_potx = 0.0
        for gauss in self.components:
            total_potx += gauss.potential(x_val)
        return total_potx

    def _get_means(self):
        """
        Get the means of the Gaussian components.

        :return: the mean vectors
        """
        means = []
        for gauss in self.components:
            means.append(gauss.get_mean())
        return means

    def _get_covs(self):
        """
        Get the covariance matrices of the Gaussian components.

        :return: the covariance matrices
        :rtype: np.ndarray list
        """
        covs = []
        for gauss in self.components:
            covs.append(gauss.get_cov())
        return covs

    def _get_log_weights(self):
        """
        Get the log weights of the Gaussian mixture components.

        :return: the log weights
        :rtype: float
        """
        log_weights = []
        for gauss in self.components:
            log_weights.append(gauss.get_log_weight())
        return log_weights

    def _get_weights(self):
        """
        Get the weights of the Gaussian mixture components.

        :return: the log weights
        :rtype: float list
        """
        weights = []
        for gauss in self.components:
            weights.append(gauss.get_weight())
        return weights

    def get_log_weight(self):
        """
        Get the total log weight of the Gaussian mixture.
        :return: The log weight
        :rtype: float
        """
        return special.logsumexp(self._get_log_weights())

    def normalize(self):
        """
        normalize the Gaussian mixture.

        :return: The normalized factor.
        :rtype: GaussianMixture
        """
        unnormalized_log_weight = self.get_log_weight()
        new_components = []
        for gauss in self.components:
            gauss_copy = gauss.copy()
            gauss_copy._add_log_weight(-1.0*unnormalized_log_weight)
            new_components.append(gauss_copy)
        return GaussianMixture(new_components)

    @property
    def is_vacuous(self):
        # TODO: see how this is used. Should this be true if there is one vacuous component? Or only if all components
        #  are vacuous? Maybe make a contains vacuous function as well.
        for gauss in self.components:
            if not gauss._is_vacuous:
                return False
        return True

    def sample(self, num_samples):
        """
        Sample from this Gaussian mixture

        :param num_samples: The number of sample to draw.
        :type num_samples: int
        :return: The samples
        :rtype: float list
        """
        weights = self._get_weights()
        component_choice_samples = np.random.choice(range(len(weights)), size=num_samples,  p=weights / np.sum(weights))
        samples = []
        for comp_index in component_choice_samples:
            samples.append(self.components[comp_index].sample(1)[0])
        return np.array(samples)

    def _get_sensible_xlim(self):
        """
        Helper function for plot function to get the x limits that contain the majority of the Gaussian mass.

        :return: The x limits.
        :rtype: float list
        """
        x_lower_candidates = []
        x_upper_candidates = []
        for gauss in self.components:
            stddev = np.sqrt(gauss.get_cov()[0, 0])
            x_lower_candidates.append(gauss.get_mean()[0, 0] - 3 * stddev)
            x_upper_candidates.append(gauss.get_mean()[0, 0] + 3 * stddev)
        x_lower = min(x_lower_candidates)
        x_upper = max(x_upper_candidates)
        return [x_lower, x_upper]

    def plot(self, log=False, xlim=None, ylim=None, show_individual_components=False):
        """
        Plot the Gaussian mixture potential function (only for 1d and 2d functions).

        :param log: if this is True, the log-potential will be plotted
        :type log: bool
        :param xlim: the x limits to plot the function over (optional)
        :type xlim: 2 element float list
        :param ylim: the y limits to plot the function over (optional and only used in 2d case)
        :type ylim: 2 element float list
        """
        if self.dim == 1:
            if xlim is None:
                xlim = self._get_sensible_xlim()
            if xlim is not None:
                x_lower = xlim[0]
                x_upper = xlim[1]
            num_points = 200
            x_series = np.linspace(x_lower, x_upper, num_points)
            total_potx = np.zeros(num_points)
            for gauss in self.components:
                if log:
                    potx = np.array([gauss.log_potential(xi) for xi in x_series])
                else:
                    potx = np.array([gauss.potential(xi) for xi in x_series])
                if show_individual_components:
                    plt.plot(x_series, potx)
                total_potx += potx
            plt.plot(x_series, total_potx)
        elif self.dim == 2:
            self._plot_2d(log=log, xlim=xlim, ylim=ylim)
        else:
            raise NotImplementedError('Plotting not implemented for dim!=1.')

    def show(self):
        """
        Print the parameters of the Gaussian mixture distribution
        """
        for i, gauss in enumerate(self.components):
            print('component ', i, '/', len(self.components))
            gauss.show()

    # pylint: disable=invalid-name
    def moment_match_to_single_gaussian(self):
        """
        Calculate the mean and covariance of the Gaussian mixture and return a Gaussian with these parameters as an
        approximation of the Gaussian Mixture.

        :return: The Gaussian approximation.
        :rtype: Gaussian
        """
        new_mean = np.zeros([self.dim, 1])
        log_weights = []
        for gauss in self.components:
            wm = gauss.get_weight() * gauss.get_mean()
            new_mean += wm
            log_weights.append(gauss.get_log_weight())
        log_sum_weights = special.logsumexp(log_weights)
        sum_weights = np.exp(log_sum_weights)
        new_mean = new_mean / sum_weights
        new_cov = np.zeros([self.dim, self.dim])
        for gauss in self.components:
            ududT = gauss.get_weight() * (gauss.get_mean() - new_mean).dot((gauss.get_mean() - new_mean).transpose())
            new_cov += gauss.get_weight() * gauss.get_cov() + ududT
        new_cov = new_cov / sum_weights
        return Gaussian(cov=new_cov, mean=new_mean, log_weight=log_sum_weights,
                        var_names=self.components[0].var_names)
    # pylint: enable=invalid-name

    def argmax(self):
        """
        Find the input vector that maximises the potential function of the Gaussian mixture.

        :return: The argmax assignment.
        :rtype: numpy.ndarray
        """
        global_maximum_potential = -float('inf')
        global_argmax = None
        success = False

        def neg_gmm_pot(x_val):
            return -1.0 * self.potential(x_val)

        for gauss in self.components:
            res = minimize(neg_gmm_pot, x0=gauss.get_mean(), method='BFGS', options={'disp': False})
            x_local_max = res.x
            if res.success:
                success = True
                local_maximum_potential = self.potential(x_local_max)
                if local_maximum_potential > global_maximum_potential:
                    global_maximum_potential = local_maximum_potential
                    global_argmax = x_local_max
        if not success:
            raise Exception('could not find optimum')
        return global_argmax

    def _argmin(self):
        """
        Find the input vector that minimises the potential function of the Gaussian mixture with negative definite
        precision matrices.

        :return: The point where the function has a global minimum.
        :rtype: np.ndarray
        """
        global_minimum_potential = float('inf')
        global_argmin = None
        success = False

        def gmm_pot(x_vals):
            return self.potential(x_vals)

        for gauss in self.components:
            res = minimize(gmm_pot, x0=gauss.get_mean(), method='BFGS', options={'disp': False})
            x_local_min = res.x
            if res.success:
                success = True
                local_minimum_potential = self.potential(x_local_min)
                if local_minimum_potential < global_minimum_potential:
                    global_minimum_potential = local_minimum_potential
                    global_argmin = x_local_min
        if not success:
            raise Exception('could not find optimum')
        return global_argmin

    # pylint: disable=invalid-name
    def _moment_match_complex_gaussian(self):
        """
        Calculate the mean and covariance of the Gaussian mixture and return a Gaussian
        with these parameters a Gaussian approximation of the Gaussian Mixture.

        :return: The Gaussian approximation.
        :rtype: Gaussian
        """
        # TODO: check if the normal moment matching function can be replaced with this one.
        new_mean = np.zeros([self.dim, 1]).astype(complex)
        weights = []
        for gauss in self.components:
            c_weight = gauss.get_complex_weight()
            c_mean = np.linalg.inv(gauss.get_K()).dot(gauss.get_h()).astype(complex)
            new_mean += c_weight * c_mean
            weights.append(c_weight)
        sum_weights = sum(weights)
        new_mean = new_mean / sum_weights
        new_cov = np.zeros([self.dim, self.dim]).astype(complex)
        for gauss in self.components:
            c_weight = gauss.get_complex_weight()
            c_mean = np.linalg.inv(gauss.get_K()).dot(gauss.get_h()).astype(complex)
            ududT = c_weight * (c_mean - new_mean).dot((c_mean - new_mean).transpose())
            c_cov = np.linalg.inv(gauss.get_K()).astype(complex)
            new_cov += c_weight * c_cov + ududT
        new_cov = new_cov / sum_weights

        new_log_weight = np.log(sum_weights)

        new_K, new_h, new_g = GaussianMixture._complex_cov_form_to_real_canform(new_cov, new_mean, new_log_weight)
        return Gaussian(K=new_K, h=new_h, g=new_g,
                        var_names=self.components[0].var_names)
    # pylint: enable=invalid-name

    @staticmethod
    # pylint: disable=invalid-name
    def _complex_cov_form_to_real_canform(cov, mean, log_weight):
        """
        A helper function for _moment_match_complex_gaussian
        :param cov: the (possibly complex) covariance matrix
        :param mean: the (possibly complex) covariance matrix
        :param log_weight: the (possibly complex) log weight
        :return: the real parts of the converted canonical parameters (K, h, g)
        """
        K = np.linalg.inv(cov)
        h = K.dot(mean)
        uTKu = ((mean.transpose()).dot(K)).dot(mean)
        log_det_term = cmath.log(np.linalg.det(2.0 * np.pi * cov))
        g = abs(log_weight - 0.5 * uTKu - 0.5 * log_det_term)
        return K.real, h.real, g.real
    # pylint: enable=invalid-name

    @staticmethod
    # pylint: disable=invalid-name
    # pylint: disable=protected-access
    def _get_inverse_gaussians(gaussian_mixture_a, gaussian_mixture_b):
        """
        A helper function for _gm_division_m1. Returns the inverse components.
        For example:
        if gaussian_mixture_a = Ga1 + Ga2
        and gaussian_mixture_b = Gb1 + Gb2
        Then return Gb1/Ga1 + Gb1/Ga2 + Gb2/Ga1 + Gb2/Ga2
        :param gaussian_mixture_a: The numerator mixture
        :param gaussian_mixture_b: The denominator mixture
        :return: The inverse Gaussian components and the mode locations of the quotient:
                 gaussian_mixture_a/gaussian_mixture_b
        """
        minimum_locations = []
        for g_a in gaussian_mixture_a.components:
            inverse_components = []
            inverse_gaussian_mixtures = []
            for g_b in gaussian_mixture_b.components:
                inverse_components.append(g_b.divide(g_a))
            inverse_gaussian_mixture = GaussianMixture(inverse_components)
            inverse_gaussian_mixtures.append(inverse_gaussian_mixture)

            minimum_loc = inverse_gaussian_mixture._argmin()
            minimum_locations.append(minimum_loc)

        def neg_gm_log_quotient_potential(x):
            """
            Get the negative of the log value of the Gaussian mixture quotient (gma/gmb)
            """
            potential = -1.0 * (gaussian_mixture_a.log_potential(x) - gaussian_mixture_b.log_potential(x))
            return potential

        mode_locations = []
        for minimum_loc in minimum_locations:
            res = minimize(neg_gm_log_quotient_potential, x0=minimum_loc, method='BFGS', options={'disp': False})
            x_local_min = res.x
            if res.success:
                mode_locations.append(x_local_min)
        unique_mode_locations = _factor_utils.remove_duplicate_values(mode_locations, tol=1e-3)
        return inverse_gaussian_mixtures, unique_mode_locations
    # pylint: enable=protected-access
    # pylint: enable=invalid-name

    @staticmethod
    # pylint: disable=invalid-name
    # pylint: disable=protected-access
    def _gm_division_m1(gaussian_mixture_a, gaussian_mixture_b):
        """
        Method 1 for approximating the quotient (gma/gmb) of two Gaussian mixtures as a Gaussian mixture.
        This is an implementation of the method described in 'A Probabilistic Graphical Model Approach to Multiple Object Tracking' (section 8.2)
        The Gaussian mixture quotient is optimised from a set of starting points in order to find modes. Laplaces method is then used at these modes
        to appromate the mode as a Gaussian. The quotient is then approximated as a sum of these Gaussians.

        :params gma: The numerator Gaussian mixture
        :params gmb: The denominator Gaussian mixture
        returns: an approximation of the quotient function as a Gaussian mixture
        """

        resulting_gaussian_components = []
        inverse_gaussian_mixtures, unique_mode_locations = GaussianMixture._get_inverse_gaussians(gaussian_mixture_a,
                                                                                                  gaussian_mixture_b)
        inv_gm = GaussianMixture(inverse_gaussian_mixtures)

        for mode_location in unique_mode_locations:
            K_i = nd.Hessian(inv_gm.log_potential)(mode_location)
            mean_i = _factor_utils.make_column_vector(mode_location)
            h_i = K_i.dot(mean_i)

            uTKu = ((mean_i.transpose()).dot(K_i)).dot(mean_i)
            log_weight_i = -0.5 * np.log(np.linalg.det(K_i / (2.0 * np.pi))) * inv_gm.log_potential(mean_i)
            g_i = log_weight_i - 0.5 * uTKu + 0.5 * np.log(np.linalg.det(K_i / (2.0 * np.pi)))

            component = Gaussian(K=K_i, h=h_i, g=g_i,
                                 var_names=gaussian_mixture_a.components[0].var_names)
            resulting_gaussian_components.append(component)
        return GaussianMixture(resulting_gaussian_components)
    # pylint: enable=protected-access
    # pylint: enable=invalid-name

    @staticmethod
    # pylint: disable=protected-access
    def _gm_division_m2(gma, gmb):
        """
        Method 2 for approximating the quotient (gma/gmb) of two Gaussian mixtures as a Gaussian mixture.
        This function does applies the moment matching equations to sets of negative definite mixtures (using
        _moment_match_complex_gaussian). For each of these mixtures, a single negative definite 'Gaussian' (h(x)) is then
        obtained. This function is then inverted (g(x) = 1/h(x)), resulting in a positive definite Gaussian function.
        The Gaussian mixture quotient (gma/gmb) is then approximated as the sum of the g(x) functions.

        This method is can be much faster than _gm_division_m1 (~10 times [tested on for one dimensional functions]),
        but also results a Gaussian mixture with the same number of components as gma. Whereas _gm_division_m2 should
        have the same number of components as the modes of the quotient function.

        :params gma: The numerator Gaussian mixture
        :params gmb: The denominator Gaussian mixture
        returns: an approximation of the quotient function as a Gaussian mixture
        """
        # TODO: add check for dimensions that are not devided by
        #       - the variances in these will not change (if this makes sense)

        inverse_gaussian_mixtures = []
        resulting_gaussian_components = []
        count = 0
        for gaussian_a in gma.components:
            inverse_components = []
            for gaussian_b in gmb.components:
                conditional = gaussian_b.divide(gaussian_a)
                inverse_components.append(conditional)
            inverse_gaussian_mixture = GaussianMixture(inverse_components)
            inverse_gaussian_mixtures.append(inverse_gaussian_mixture)
            inverse_gaussian_approx = inverse_gaussian_mixture._moment_match_complex_gaussian()
            if _factor_utils.is_pos_def(inverse_gaussian_approx.K):
                count += 1
                # raise ValueError(' precision is negative definite.')
                print('Warning: precision is negative definite.')
            gaussian_approx = inverse_gaussian_approx._invert()
            resulting_gaussian_components.append(gaussian_approx)
        print('num pos def = ', count, '/', len(gma.components))
        resulting_gm = GaussianMixture(resulting_gaussian_components)
        return resulting_gm
    # pylint: enable=protected-access

    def _get_limits_for_2d_plot(self):  # pragma: no cover
        """
        Get x and y limits which includes most of the Gaussian mixture's mass, by considering
        the mean and variance of each Gaussian component.
        """
        x_lower_candidates = []
        x_upper_candidates = []
        y_lower_candidates = []
        y_upper_candidates = []
        for gauss in self.components:
            stddev_x = np.sqrt(gauss.get_cov()[0, 0])
            stddev_y = np.sqrt(gauss.get_cov()[1, 1])
            mean_x = gauss.get_mean()[0, 0]
            mean_y = gauss.get_mean()[1, 0]
            x_lower_candidates.append(mean_x - 3.0 * stddev_x)
            x_upper_candidates.append(mean_x + 3.0 * stddev_x)
            y_lower_candidates.append(mean_y - 3.0 * stddev_y)
            y_upper_candidates.append(mean_y + 3.0 * stddev_y)
        x_lower = min(x_lower_candidates)
        x_upper = max(x_upper_candidates)
        y_lower = min(y_lower_candidates)
        y_upper = max(y_upper_candidates)
        return [x_lower, x_upper], [y_lower, y_upper]

    def _plot_2d(self, log, xlim, ylim):  # pragma: no cover
        """
        Plot a 2d Gaussian mixture potential function
        :param log: if this is True, the log-potential will be plotted
        :param xlim: the x limits to plot the function over (optional)
        :param ylim: the y limits to plot the function over (optional)
        """
        if xlim is None and ylim is None:
            xlim, ylim = self._get_limits_for_2d_plot()
        elif xlim is not None and ylim is not None:
            pass
        else:
            print('Warning: partial limits received. Generating limits automatically.')
            xlim, ylim = self._get_limits_for_2d_plot()

        xlabel = self.var_names[0]
        ylabel = self.var_names[1]

        if not log:
            _factor_utils.plot_2d(func=self.potential, xlim=xlim, ylim=ylim, xlabel=xlabel, ylabel=ylabel)
        else:
            _factor_utils.plot_2d(func=self.log_potential, xlim=xlim, ylim=ylim, xlabel=xlabel, ylabel=ylabel)
